from collections.abc import Iterable, Iterator, Sequence
from functools import cached_property
from typing import TYPE_CHECKING, Any

from eth_pydantic_types import HexBytes, HexStr
from eth_utils import encode_hex, is_hex, keccak, to_hex
from ethpm_types.abi import EventABI
from pydantic import BaseModel, field_serializer, field_validator, model_validator
from web3.types import FilterParams

from ape.exceptions import ContractNotFoundError
from ape.types.address import AddressType
from ape.types.basic import HexInt
from ape.utils.abi import LogInputABICollection, encode_topics
from ape.utils.basemodel import BaseInterfaceModel, ExtraAttributesMixin, ExtraModelAttributes
from ape.utils.misc import ZERO_ADDRESS, log_instead_of_fail

if TYPE_CHECKING:
    from ape.api.providers import BlockAPI
    from ape.contracts import ContractEvent


TopicFilter = Sequence[HexStr | None | Sequence[HexStr | None]]


class LogFilter(BaseModel):
    addresses: list[AddressType] = []
    events: list[EventABI] = []
    topic_filter: TopicFilter = []
    start_block: int = 0
    stop_block: int | None = None  # Use block height
    selectors: dict[str, EventABI] = {}

    @model_validator(mode="before")
    @classmethod
    def compute_selectors(cls, values):
        values["selectors"] = {
            encode_hex(keccak(text=event.selector)): event for event in values.get("events") or []
        }

        return values

    @field_validator("start_block", mode="before")
    @classmethod
    def validate_start_block(cls, value):
        return value or 0

    @field_validator("addresses", "events", "topic_filter", mode="before")
    @classmethod
    def _convert_none_to_empty_list(cls, value):
        return value or []

    @field_validator("selectors", mode="before")
    @classmethod
    def _convert_none_to_dict(cls, value):
        return value or {}

    def model_dump(self, *args, **kwargs):
        return FilterParams(
            address=self.addresses,
            fromBlock=to_hex(self.start_block),
            toBlock=to_hex(self.stop_block or self.start_block),
            topics=self.topic_filter,  # type: ignore
        )

    @classmethod
    def from_event(
        cls,
        event: "EventABI | ContractEvent",
        search_topics: dict[str, Any] | None = None,
        addresses: list[AddressType] | None = None,
        start_block=None,
        stop_block=None,
    ):
        """
        Construct a log filter from an event topic query.
        """
        abi = getattr(event, "abi", event)
        topic_filter = encode_topics(abi, search_topics or {})
        return cls(
            addresses=addresses or [],
            events=[abi],
            topic_filter=topic_filter,
            start_block=start_block,
            stop_block=stop_block,
        )


class BaseContractLog(BaseInterfaceModel):
    """
    Base class representing information relevant to an event instance.
    """

    event_name: str
    """The name of the event."""

    contract_address: AddressType = ZERO_ADDRESS
    """The contract responsible for emitting the log."""

    event_arguments: dict[str, Any] = {}
    """The arguments to the event, including both indexed and non-indexed data."""

    def __eq__(self, other: Any) -> bool:
        if self.contract_address != other.contract_address or self.event_name != other.event_name:
            return False

        for k, v in self.event_arguments.items():
            other_v = other.event_arguments.get(k)
            if v != other_v:
                return False

        return True

    @field_serializer("event_arguments")
    def _serialize_event_arguments(self, event_arguments, info):
        """
        Because of an issue with BigInt in Pydantic,
        (https://github.com/pydantic/pydantic/issues/10152)
        we have to ensure these are regular ints.
        """
        return self._serialize_value(event_arguments, info)

    def _serialize_value(self, value: Any, info) -> Any:
        if isinstance(value, int):
            # Handle custom ints.
            return int(value)

        elif isinstance(value, HexBytes):
            return to_hex(value) if info.mode == "json" else value

        elif isinstance(value, str):
            # Avoiding str triggering iterable condition.
            return value

        elif isinstance(value, dict):
            # Also, avoid handling dict in the iterable case.
            return {k: self._serialize_value(v, info) for k, v in value.items()}

        elif isinstance(value, Iterable):
            return [self._serialize_value(v, info) for v in value]

        return value


# TODO: In 0.9, make this correctly serialize back to log data you get from RPCs.
class ContractLog(ExtraAttributesMixin, BaseContractLog):
    """
    An instance of a log from a contract.
    """

    def __init__(self, *args, **kwargs):
        abi = kwargs.pop("_abi", None)
        super().__init__(*args, **kwargs)
        if isinstance(abi, LogInputABICollection):
            abi = abi.abi

        self._abi = abi

    transaction_hash: Any
    """The hash of the transaction containing this log."""

    block_number: HexInt
    """The number of the block containing the transaction that produced this log."""

    block_hash: Any
    """The hash of the block containing the transaction that produced this log."""

    log_index: HexInt
    """The index of the log on the transaction."""

    transaction_index: HexInt | None = None
    """
    The index of the transaction's position when the log was created.
    Is `None` when from the pending block.
    """

    removed: bool = False
    """
    ``True`` if this log has been removed.
    """

    @field_validator("transaction_hash", mode="before")
    @classmethod
    def _validate_transaction_hash(cls, transaction_hash):
        if (
            isinstance(transaction_hash, str)
            and is_hex(transaction_hash)
            and not transaction_hash.startswith("0x")
        ):
            return f"0x{transaction_hash}"

        return transaction_hash

    @field_serializer("transaction_hash", "block_hash")
    def _serialize_hashes(self, value, info):
        return self._serialize_value(value, info)

    # NOTE: This class has an overridden `__getattr__` method, but `block` is a reserved keyword
    #       in most smart contract languages, so it is safe to use. Purposely avoid adding
    #       `.datetime` and `.timestamp` in case they are used as event arg names.
    @cached_property
    def block(self) -> "BlockAPI":
        return self.chain_manager.blocks[self.block_number]

    @property
    def abi(self) -> EventABI:
        """
        An ABI describing the event.
        """
        if abi := self._abi:
            return abi

        try:
            contract = self.chain_manager.contracts[self.contract_address]
        except (ContractNotFoundError, KeyError):
            pass

        else:
            for event_abi in contract.events:
                if event_abi.name == self.event_name and len(event_abi.inputs) == len(
                    self.event_arguments
                ):
                    abi = event_abi

        if abi is None:
            # Last case scenario, try to calculate it.
            # NOTE: This is a rare edge case that shouldn't really happen,
            #   so this is a lower priority.
            # TODO: Handle inputs here.
            abi = EventABI(name=self.event_name)

        self._abi = abi
        return abi

    @cached_property
    def topics(self) -> list[HexStr]:
        """
        The encoded hex-str topics values.
        """
        return encode_topics(self.abi, self.event_arguments)

    @property
    def timestamp(self) -> int:
        """
        The UNIX timestamp of when the event was emitted.

        NOTE: This performs a block lookup.
        """
        return self.block.timestamp

    @property
    def _event_args_str(self) -> str:
        return " ".join(f"{key}={val}" for key, val in self.event_arguments.items())

    def __str__(self) -> str:
        return f"{self.event_name}({self._event_args_str})"

    @log_instead_of_fail(default="<ContractLog>")
    def __repr__(self) -> str:
        event_arg_str = self._event_args_str
        suffix = f" {event_arg_str}" if event_arg_str else ""
        return f"<{self.event_name or 'Event'}{suffix}>"

    def __ape_extra_attributes__(self) -> Iterator[ExtraModelAttributes]:
        yield ExtraModelAttributes(
            name=self.event_name or "event",
            attributes=lambda: self.event_arguments or {},
            include_getattr=True,
            include_getitem=True,
        )

    def __contains__(self, item: str) -> bool:
        return item in self.event_arguments

    def __eq__(self, other: Any) -> bool:
        """
        Check for equality between this instance and another ContractLog instance.

        If the other object is not an instance of ContractLog, this method returns
        NotImplemented. This triggers the Python interpreter to call the __eq__ method
        on the other object (i.e., y.__eq__(x)) if it is defined, allowing for a custom
        comparison. This behavior is leveraged by the MockContractLog class to handle
        custom comparison logic between ContractLog and MockContractLog instances.

        Args:
            other (Any): The object to compare with this instance.

        Returns:
            bool: True if the two instances are equal, False otherwise.
        """

        if not isinstance(other, ContractLog):
            return NotImplemented

        # call __eq__ on parent class
        return super().__eq__(other)

    def get(self, item: str, default: Any | None = None) -> Any:
        return self.event_arguments.get(item, default)


def _equal_event_inputs(mock_input: Any, real_input: Any) -> bool:
    if mock_input is None:
        # Check is skipped.
        return True

    elif isinstance(mock_input, (list, tuple)):
        if not isinstance(real_input, (list, tuple)) or len(real_input) != len(mock_input):
            return False

        return all(_equal_event_inputs(m, r) for m, r in zip(mock_input, real_input, strict=True))

    else:
        return mock_input == real_input


class MockContractLog(BaseContractLog):
    """
    A mock version of the ContractLog class used for testing purposes.
    This class is designed to match a subset of event arguments in a ContractLog instance
    by only comparing those event arguments that the user explicitly provides.

    Inherits from :class:`~ape.types.BaseContractLog`, and overrides the
    equality method for custom comparison
    of event arguments between a MockContractLog and a ContractLog instance.
    """

    def __eq__(self, other: Any) -> bool:
        if (
            not hasattr(other, "contract_address")
            or not hasattr(other, "event_name")
            or self.contract_address != other.contract_address
            or self.event_name != other.event_name
        ):
            return False

        # NOTE: `self.event_arguments` contains a subset of items from `other.event_arguments`,
        #       but we skip those the user doesn't care to check
        for name, value in self.event_arguments.items():
            other_input = other.event_arguments.get(name)
            if not _equal_event_inputs(value, other_input):
                # Only exit on False; Else, keep checking.
                return False

        return True


class ContractLogContainer(list):
    """
    Container for ContractLogs which is adding capability of filtering logs
    """

    def filter(self, event: "ContractEvent", **kwargs) -> list[ContractLog]:
        return [
            x
            for x in self
            if x.event_name == event.name
            and x.contract_address == event.contract
            and all(v == x.event_arguments.get(k) and v is not None for k, v in kwargs.items())
        ]

    def __contains__(self, val: Any) -> bool:
        return any(log == val for log in self)



from datetime import datetime, timedelta, timezone

import pytest

from ape.exceptions import ConversionError


@pytest.mark.parametrize(
    "args",
    (
        [
            "10-15-2021 12:15:12",
            "2021-10-15 12:15:12",
            "15-10-2021 12:15:12",
            datetime.fromisoformat("2021-10-15 12:15:12"),
        ]
    ),
)
def test_convert_string_timestamp(args, convert):
    assert convert(args, int) == 1634300112


@pytest.mark.parametrize(
    "args",
    (
        [
            "100.0",
            "foobar",
            "2001-01-01 12:15:12 123",
        ]
    ),
)
def test_convert_string_bad_timestamp(args, convert):
    with pytest.raises(ConversionError):
        convert(args, int)


def test_convert_timedelta_timestamp(convert):
    delta = timedelta(days=2, hours=2, minutes=12)
    assert convert(delta, int) == int((datetime.now(timezone.utc) + delta).timestamp())



"""
Warnings are great for Pytest because they always show up
at the end of a test-run.
"""

import warnings


class InvalidIsolationWarning(Warning):
    """
    Occurs when fixtures disrupt isolation causing performance degradation.
    """


def warn_invalid_isolation():
    message = (
        "Invalid isolation; Ensure session|package|module|class scoped fixtures "
        "run earlier. Rebasing fixtures is costly."
    )
    warnings.warn(message, InvalidIsolationWarning)



from contextlib import contextmanager
from pathlib import Path

from ape.managers.base import BaseManager
from ape.types.address import AddressType
from ape.utils.basemodel import BaseModel, DiskCacheableModel
from ape.utils.os import create_tempdir


class Deployment(BaseModel):
    """
    A single deployment of a contract in a network.
    """

    address: AddressType
    transaction_hash: str | None = None

    def __getitem__(self, key: str):
        # Mainly exists for backwards compat.
        if key == "address":
            return self.address
        elif key == "transaction_hash":
            return self.transaction_hash

        raise KeyError(key)

    def get(self, key: str):
        # Mainly exists for backwards compat.
        try:
            return self[key]
        except KeyError:
            return None


class Deployments(DiskCacheableModel):
    """The deployments structured JSON."""

    ecosystems: dict[str, dict[str, dict[str, list[Deployment]]]] = {}


class DeploymentDiskCache(BaseManager):
    """
    Manage cached contract deployments.
    """

    def __init__(self):
        # NOTE: For some reason, deployments are all inside their ecosystem folders,
        #   but they still have the ecosystem key. Hence, the weird structure here.
        self._deployments: dict[str, Deployments] = {}
        self._base_path = None

    @property
    def _is_live_network(self) -> bool:
        return bool(self.network_manager.active_provider) and not self.provider.network.is_dev

    @property
    def cachefile(self) -> Path:
        base_path = self._base_path or self.provider.network.ecosystem.data_folder
        return base_path / "deployments_map.json"

    @property
    def _all_deployments(self) -> Deployments:
        if not self._is_live_network:
            # No file.
            if "local" not in self._deployments:
                self._deployments["local"] = Deployments()

            return self._deployments["local"]

        ecosystem_name = self.provider.network.ecosystem.name
        if ecosystem_name not in self._deployments:
            self._deployments[ecosystem_name] = Deployments.model_validate_file(self.cachefile)

        return self._deployments[ecosystem_name]

    def __getitem__(self, contract_name: str) -> list[Deployment]:
        return self.get_deployments(contract_name)

    def __setitem__(self, contract_name, deployments: list[Deployment]):
        self._set_deployments(contract_name, deployments)

    def __delitem__(self, contract_name: str):
        self.remove_deployments(contract_name)

    def __contains__(self, contract_name: str):
        return bool(self.get_deployments(contract_name))

    def get_deployments(
        self,
        contract_name: str,
        ecosystem_key: str | None = None,
        network_key: str | None = None,
    ) -> list[Deployment]:
        """
        Get the deployments of the given contract on the currently connected network.

        Args:
            contract_name (str): The name of the deployed contract.
            ecosystem_key (str | None): The ecosystem key. Defaults to
              the connected ecosystem's name.
            network_key (str | None): The network key. Defaults to the
              connected network's name.

        Returns:
            list[Deployment]
        """
        if not self.network_manager.connected and (not ecosystem_key or not network_key):
            # Allows it to work when not connected (testing?)
            return []

        ecosystem_name = ecosystem_key or self.provider.network.ecosystem.name
        network_name = network_key or self.provider.network.name.replace("-fork", "")
        return (
            self._all_deployments.ecosystems.get(ecosystem_name, {})
            .get(network_name, {})
            .get(contract_name, [])
        )

    def cache_deployment(
        self,
        address: AddressType,
        contract_name: str,
        transaction_hash: str | None = None,
        ecosystem_key: str | None = None,
        network_key: str | None = None,
    ):
        """
        Update the deployments cache with a new contract.

        Args:
            address (AddressType): The address of the contract.
            contract_name (str): The name of the contract type.
            transaction_hash (Optional[str]): Optionally, the transaction has
              associated with the deployment transaction.
            ecosystem_key (str | None): The ecosystem key. Defaults to
              the connected ecosystem's name.
            network_key (str | None): The network key. Defaults to the
              connected network's name.
        """
        deployments = [
            *self.get_deployments(contract_name),
            Deployment(address=address, transaction_hash=transaction_hash),
        ]
        self._set_deployments(
            contract_name,
            deployments,
            ecosystem_key=ecosystem_key,
            network_key=network_key,
        )

    @contextmanager
    def use_temporary_cache(self):
        base_path = self._base_path
        deployments = self._deployments
        with create_tempdir() as temp_path:
            self._base_path = temp_path
            self._deployments = {}
            yield

        self._base_path = base_path
        self._deployments = deployments

    def _set_deployments(
        self,
        contract_name: str,
        deployments: list[Deployment],
        ecosystem_key: str | None = None,
        network_key: str | None = None,
    ):
        ecosystem_name = ecosystem_key or self.provider.network.ecosystem.name
        network_name = network_key or self.provider.network.name.replace("-fork", "")
        self._all_deployments.ecosystems.setdefault(ecosystem_name, {})
        self._all_deployments.ecosystems[ecosystem_name].setdefault(network_name, {})
        self._all_deployments.ecosystems[ecosystem_name][network_name][contract_name] = deployments

        # For live networks, cache the deployments to a file as well.
        if self._is_live_network and ecosystem_name in self._deployments:
            self._deployments[ecosystem_name].model_dump_file()

    def remove_deployments(self, contract_name: str):
        self._set_deployments(contract_name, [])

    def clear_local(self):
        self._deployments["local"] = Deployments()



import os
import subprocess
from pathlib import Path

import pytest

from ape import __all__
from tests.conftest import ApeSubprocessRunner
from tests.integration.cli.utils import skip_projects, skip_projects_except


@pytest.fixture
def console_runner(config):
    class ConsoleSubprocessRunner(ApeSubprocessRunner):
        def __init__(self):
            super().__init__("console", data_folder=config.DATA_FOLDER)

    return ConsoleSubprocessRunner()


@pytest.fixture(params=("path", "root"))
def extras_base_path(integ_project, request):
    if request.param == "path":
        yield integ_project.path
    else:
        yield integ_project.config_manager.DATA_FOLDER


# Simple single namespace example
EXTRAS_SCRIPT_1 = """
A = 1
def a():
    return A
"""

# Tests ape_init_extras function alters namespace
EXTRAS_SCRIPT_2 = """
A = 1

def ape_init_extras():
    global A
    A = 2
"""

# Tests that namespace kwargs are available
EXTRAS_SCRIPT_3 = """
def ape_init_extras(project):
    assert project
"""

# Tests that returned dict is added to namespace
EXTRAS_SCRIPT_4 = """
B = 4

def ape_init_extras():
    return {"A": 1, "B": 2}
"""

# Tests that we can import a local package
EXTRAS_SCRIPT_5 = """
from dependency_in_project_only.importme import import_me

import_me()
"""


def no_console_error(result):
    return (
        "NameError" not in result.output
        and "AssertionError" not in result.output
        and "ModuleNotFoundError" not in result.output
    )


def write_ape_console_extras(base_path, contents):
    extras_file = base_path.joinpath("ape_console_extras.py")
    extras_file.write_text(contents, encoding="utf8")
    return extras_file


@pytest.fixture(autouse=True)
def clean_console_rc_write(project):
    yield

    global_extras = project.config_manager.DATA_FOLDER.joinpath("ape_console_extras.py")
    if global_extras.is_file():
        global_extras.unlink()

    project_extras = project.path.joinpath("ape_console_extras.py")
    if project_extras.is_file():
        project_extras.unlink()


# NOTE: We export `__all__` into the IPython session that the console runs in
@skip_projects("geth")
@pytest.mark.parametrize("item", __all__)
def test_console(ape_cli, runner, item, project):
    arguments = ["console", "--project", f"{project.path}"]
    result = runner.invoke(ape_cli, arguments, input=f"{item}\nexit\n", catch_exceptions=False)
    assert result.exit_code == 0, result.output
    assert no_console_error(result), result.output
    arguments.extend(("-v", "debug"))
    result = runner.invoke(
        ape_cli,
        arguments,
        input=f"{item}\nexit\n",
        catch_exceptions=False,
    )
    assert result.exit_code == 0, result.output
    assert no_console_error(result), result.output


@skip_projects("geth")
def test_console_extras(integ_project, extras_base_path, ape_cli, runner):
    write_ape_console_extras(extras_base_path, EXTRAS_SCRIPT_1)
    arguments = ("console", "--project", f"{integ_project.path}")

    result = runner.invoke(
        ape_cli,
        arguments,
        input="\n".join(["assert A == 1", "exit"]) + "\n",
        catch_exceptions=False,
    )
    assert result.exit_code == 0, result.output
    assert no_console_error(result), result.output

    result = runner.invoke(
        ape_cli,
        arguments,
        input="\n".join(["assert a() == 1", "exit"]) + "\n",
        catch_exceptions=False,
    )
    assert result.exit_code == 0, result.output
    assert no_console_error(result), result.output


@skip_projects("geth")
def test_console_init_extras(integ_project, extras_base_path, ape_cli, runner):
    write_ape_console_extras(extras_base_path, EXTRAS_SCRIPT_2)
    arguments = ("console", "--project", f"{integ_project.path}")
    result = runner.invoke(
        ape_cli,
        arguments,
        input="print('a:', A)\nassert A == 2\nexit\n",
        catch_exceptions=False,
    )
    assert result.exit_code == 0, result.output
    assert no_console_error(result), result.output


@skip_projects("geth")
def test_console_init_extras_kwargs(integ_project, extras_base_path, ape_cli, runner):
    write_ape_console_extras(extras_base_path, EXTRAS_SCRIPT_3)
    arguments = ("console", "--project", f"{integ_project.path}")
    result = runner.invoke(ape_cli, arguments, input="exit\n", catch_exceptions=False)
    assert result.exit_code == 0, result.output
    assert no_console_error(result), result.output


@skip_projects("geth")
def test_console_init_extras_return(integ_project, extras_base_path, ape_cli, runner):
    write_ape_console_extras(extras_base_path, EXTRAS_SCRIPT_4)
    arguments = ("console", "--project", f"{integ_project.path}")

    # Test asserts returned A exists and B is not overwritten
    result = runner.invoke(
        ape_cli,
        arguments,
        input="\n".join(
            [
                "assert A == 1, 'unexpected A'",
                # symbols from ape_init_extras should apply before file namespace
                "assert B == 2, 'unexpected B'",
                "exit",
            ]
        )
        + "\n",
        catch_exceptions=False,
    )
    assert result.exit_code == 0, result.output
    assert no_console_error(result), result.output


@skip_projects_except("only-dependencies")
def test_console_import_local_path(integ_project, ape_cli, runner):
    # NOTE: Don't use temp-path! Temp-path projects do not copy Python modules.
    path = Path(__file__).parent / "projects" / "only-dependencies"
    arguments = ("console", "--project", f"{path}")
    result = runner.invoke(
        ape_cli,
        arguments,
        input="\n".join(["from dependency_in_project_only.importme import import_me", "exit"])
        + "\n",
    )
    assert result.exit_code == 0, result.output
    assert no_console_error(result), result.output


@skip_projects_except("only-dependencies")
def test_console_import_local_path_in_extras_file(integ_project, extras_base_path, ape_cli, runner):
    # NOTE: Don't use tmp path! Temp projects do not copy Python modules.
    path = Path(__file__).parent / "projects" / "only-dependencies"
    write_ape_console_extras(extras_base_path, EXTRAS_SCRIPT_5)
    arguments = ("console", "--project", f"{path}")
    result = runner.invoke(
        ape_cli,
        arguments,
        input="exit\n",
        catch_exceptions=False,
    )
    assert result.exit_code == 0, result.output
    assert no_console_error(result), result.output


@skip_projects_except("only-dependencies")
def test_console_ape_magic(integ_project, ape_cli, runner):
    arguments = ("console", "--project", f"{integ_project.path}")
    result = runner.invoke(
        ape_cli,
        arguments,
        input="%load_ext ape_console.plugin\n%ape--help\nexit\n",
        catch_exceptions=False,
    )
    assert result.exit_code == 0, result.output
    assert no_console_error(result), result.output


@skip_projects_except("only-dependencies")
def test_console_bal_magic(integ_project, ape_cli, runner, keyfile_account):
    arguments = ("console", "--project", f"{integ_project.path}")
    cases = (
        "%load_ext ape_console.plugin",
        "%bal acct",
        "%bal acct.alias",
        "%bal acct.address",
        "%bal int(acct.address, 16)",
    )
    cmd_ls = [f"acct = accounts.load('{keyfile_account.alias}')", *cases, "exit"]
    cmd_str = "\n".join(cmd_ls)
    result = runner.invoke(
        ape_cli,
        arguments,
        input=f"{cmd_str}\n",
        catch_exceptions=False,
    )
    assert result.exit_code == 0, result.output
    assert no_console_error(result), result.output


@skip_projects_except("with-contracts")
def test_uncaught_txn_err(integ_project, console_runner):
    cmd_ls = [
        "%load_ext ape_console.plugin",
        "account = accounts.test_accounts[0]",
        "contract = account.deploy(project.ContractA)",
        "contract.setNumber(5, sender=account)",
        "exit",
    ]
    cmd_str = "\n".join(cmd_ls)
    console_runner.project = integ_project
    result = console_runner.invoke(input=f"{cmd_str}\n")
    assert "ERROR:    (ContractLogicError) Transaction failed." in result.output


def test_console_none_network(integ_project, ape_cli, runner):
    arguments = ("console", "--project", f"{integ_project.path}", "--network", "None")
    result = runner.invoke(ape_cli, arguments, input="exit\n", catch_exceptions=False)
    assert result.exit_code == 0


@skip_projects_except("with-contracts")
def test_console_natspecs(integ_project, project, console_runner):
    """
    This test shows that the various natspec integrations with ABI-backed
    types work in ``ape console``.
    """
    contract_code = project.SolidityContract.contract_type.model_dump_json(by_alias=True)
    # flake8: noqa
    cmd_ls = [
        "%load_ext ape_console.plugin",
        f"contract_container = compilers.ethpm.compile_code('{contract_code}')",
        "account = accounts.test_accounts[0]",
        "contract = account.deploy(contract_container, 123)",
        "print('0: method')",
        "contract.setNumber",
        "print('1: event')",
        "contract.NumberChange",
        "print('2: error')",
        "contract.ACustomError",
        "exit",
    ]
    cmd_str = "\n".join(cmd_ls)
    expected_method = """
setNumber(uint256 num)
  @custom:emits Emits a `NumberChange` event with the previous number, the new number, and the previous block hash
  @custom:modifies Sets the `myNumber` state variable
  @custom:require num Must not be equal to 5
  @details Only the owner can call this function. The new number cannot be 5.
  @param num uint256 The new number to be set
""".strip()
    expected_event = """
NumberChange(bytes32 b, uint256 prevNum, string dynData, uint256 indexed newNum, string indexed dynIndexed)
  @details Emitted when number is changed. `newNum` is the new number from the call. Expected every time number changes.
""".strip()
    # flake8: on
    result = console_runner.invoke("--project", f"{integ_project.path}", input=f"{cmd_str}\n")

    # Getting rid of newlines as terminal-breakage never consistent in tests.
    actual = result.output.replace("\n", "")

    assert all(ln in actual for ln in expected_method.splitlines())
    assert all(ln in actual for ln in expected_event.splitlines())


@skip_projects_except("with-contracts")
def test_console_code(integ_project, mocker, console_runner):
    """
    Testing the -c | --code option.
    """
    result = console_runner.invoke(
        "--project", f"{integ_project.path}", "--code", "chain\nx = 3\nx + 1"
    )
    expected = "Out[1]: <ChainManager (id=1337)>\nOut[3]: 4\n"
    assert result.output == expected



from importlib import import_module

from ape.plugins import Config, register


@register(Config)
def config_class():
    from ape_networks.config import NetworksConfig

    return NetworksConfig


def __getattr__(name: str):
    if name in ("NetworksConfig", "CustomNetwork"):
        return getattr(import_module("ape_networks.config"), name)

    else:
        raise AttributeError(name)


__all__ = ["NetworksConfig"]



import sys

import pytest

from ape import Project
from ape.managers.accounts import AccountManager
from ape.managers.project import LocalProject
from ape.utils import ManagerAccessMixin, create_tempdir
from ape_console._cli import CONSOLE_EXTRAS_FILENAME, ApeConsoleNamespace, console
from ape_console.plugin import custom_exception_handler


@pytest.fixture(autouse=True)
def mock_console(mocker):
    """Prevent console from actually launching."""
    return mocker.patch("ape_console._cli._launch_console")


def test_console_extras_uses_ape_namespace(mocker, mock_console):
    """
    Test that if console is given extras, those are included in the console
    but not as args to the extras files, as those files expect items from the
    default ape namespace.
    """
    namespace_patch = mocker.patch("ape_console._cli._create_namespace")
    accounts_custom = mocker.MagicMock()
    extras = {"accounts": accounts_custom}
    console(extra_locals=extras)
    actual = namespace_patch.call_args[1]
    assert actual["accounts"] == accounts_custom


def test_console_custom_project(mock_console):
    with create_tempdir() as path:
        project = Project(path)
        extras_file = path / CONSOLE_EXTRAS_FILENAME
        extras_file.touch()
        console(project=project)
        extras = mock_console.call_args[0][0]
        actual = extras["project"]

    assert actual == project

    # Ensure sys.path was updated correctly.
    assert sys.path[0] == str(project.path)


def test_custom_exception_handler_handles_non_ape_project(mocker):
    """
    If the user has assigned the variable ``project`` to something else
    in their active ``ape console`` session, the exception handler
    **SHOULD NOT** attempt to use its ``.path``.
    """
    session = mocker.MagicMock()
    session.user_ns = {"project": 123}  # Like doing `project = 123` in a console.

    err = Exception()

    handler_patch = mocker.patch("ape_console.plugin.handle_ape_exception")

    # Execute - this was failing before the fix.
    custom_exception_handler(session, None, err, None)

    # We are expecting the local project's path in the handler.
    expected_path = ManagerAccessMixin.local_project.path
    handler_patch.assert_called_once_with(err, [expected_path])


class TestApeConsoleNamespace:
    def test_accounts(self):
        extras = ApeConsoleNamespace()
        assert isinstance(extras["accounts"], AccountManager)

    @pytest.mark.parametrize("scope", ("local", "global"))
    def test_extras(self, scope):
        extras = ApeConsoleNamespace()
        _ = getattr(extras, f"_{scope}_extras")
        extras.__dict__[f"_{scope}_extras"] = {"foo": "123"}
        assert extras["foo"] == "123"

    @pytest.mark.parametrize("scope", ("local", "global"))
    def test_extras_load_using_ape_namespace(self, scope):
        extras = ApeConsoleNamespace()
        _ = getattr(extras, f"_{scope}_path")
        extras_content = """
def ape_init_extras(project):
    return {"foo": type(project)}
"""
        with create_tempdir() as temp:
            extras_file = temp / CONSOLE_EXTRAS_FILENAME
            extras_file.write_text(extras_content)
            extras.__dict__[f"_{scope}_path"] = extras_file
            extras.__dict__.pop(f"_{scope}_extras", None)
            assert extras["foo"] is LocalProject



from pathlib import Path

import pytest
from requests.exceptions import ConnectTimeout, HTTPError

from ape.exceptions import ProjectError
from ape.utils._github import _GithubClient
from ape.utils.os import create_tempdir

ORG_NAME = "test"
REPO_NAME = "path"
REPO_PATH = f"{ORG_NAME}/{REPO_NAME}"


@pytest.fixture(autouse=True)
def clear_repo_cache(github_client):
    def clear():
        if REPO_PATH in github_client._repo_cache:
            del github_client._repo_cache[REPO_PATH]

    clear()
    yield
    clear()


@pytest.fixture
def mock_session(mocker):
    return mocker.MagicMock()


@pytest.fixture
def mock_release(mocker):
    release = mocker.MagicMock()
    release.json.return_value = {"name": REPO_NAME}
    return release


@pytest.fixture
def github_client(mock_session):
    return _GithubClient(session=mock_session)


class TestGithubClient:
    def test_get_repo_unknown_repo(self, mocker, mock_session):
        client = _GithubClient(session=mock_session)

        # Make it raise 404.
        response = mocker.MagicMock()
        response.status_code = 404
        error = HTTPError(response=response)
        mock_session.request.side_effect = error

        with pytest.raises(ProjectError, match=f"Unknown repository '{ORG_NAME}/{REPO_NAME}'"):
            client.get_repo(ORG_NAME, REPO_NAME)

    def test_clone_repo(self, mocker):
        # NOTE: this test actually clones the repo.
        client = _GithubClient()
        git_patch = mocker.patch("ape.utils._github.subprocess.call")
        git_patch.return_value = 0
        with create_tempdir() as temp_dir:
            try:
                client.clone_repo("dapphub", "ds-test", Path(temp_dir), branch="master")
            except ConnectTimeout:
                pytest.xfail("Internet required to run this test.")

        cmd = git_patch.call_args[0][0]
        assert cmd[0].endswith("git")
        assert cmd[1] == "-c"
        assert cmd[2] == "advice.detachedHead=false"
        assert cmd[3] == "clone"
        assert cmd[4] == "https://github.com/dapphub/ds-test.git"
        # cmd[5] is the temporary output path
        assert cmd[6] == "--branch"
        assert cmd[7] == "master"

    def test_get_release(self, github_client, mock_session):
        version = "0.1.0"
        github_client.get_release(ORG_NAME, REPO_NAME, "0.1.0")
        base_uri = f"https://api.github.com/repos/{ORG_NAME}/{REPO_NAME}/releases/tags"
        expected_uri = f"{base_uri}/{version}"
        assert mock_session.request.call_args[0] == ("GET", expected_uri)

    @pytest.mark.parametrize("version", ("0.1.0", "v0.1.0"))
    def test_get_release_retry(self, mock_release, github_client, mock_session, version):
        """
        Ensure after failing to get a release, we re-attempt with
        out a v-prefix.
        """
        opposite = version.lstrip("v") if version.startswith("v") else f"v{version}"

        def side_effect(method, uri, *arg, **kwargs):
            _version = uri.split("/")[-1]
            if _version == version:
                # Force it to try the opposite.
                raise ValueError()

            return mock_release

        mock_session.request.side_effect = side_effect
        actual = github_client.get_release(ORG_NAME, REPO_NAME, version)
        assert actual["name"] == REPO_NAME
        calls = mock_session.request.call_args_list[-2:]
        expected_uri = "https://api.github.com/repos/test/path/releases/tags"
        assert calls[0][0] == ("GET", f"{expected_uri}/{version}")
        assert calls[1][0] == ("GET", f"{expected_uri}/{opposite}")

    def test_get_org_repos(self, github_client, mock_session):
        _ = list(github_client.get_org_repos())
        call = mock_session.method_calls[-1]
        params = call.kwargs["params"]
        # Show we are fetching more than the default 30 per page.
        assert params == {"per_page": 100, "page": 1}

    def test_available_plugins(self, mocker, github_client, mock_session):
        response1 = mocker.MagicMock()
        response1.json.return_value = [{"name": "ape-myplugin"}]
        response2 = mocker.MagicMock()
        response2.json.return_value = []

        def get_org_repos(method, url, **kwargs):
            if kwargs["params"]["page"] == 1:
                return response1
            else:
                # End.
                return response2

        mock_session.request.side_effect = get_org_repos
        actual = github_client.available_plugins
        assert actual == {"ape_myplugin"}

    def test_available_plugins_handles_401(self, mocker, github_client, mock_session, ape_caplog):
        """
        When you get a 401 from using a token, Ape's GitHub client should not
        only warn the user but retry the request w/o authorization, as it likely
        will still work.
        """
        mock_session.headers = {"Authorization": "token mytoken"}

        response1 = mocker.MagicMock()
        response1.json.return_value = [{"name": "ape-myplugin"}]
        response2 = mocker.MagicMock()
        response2.json.return_value = []

        bad_auth_response = mocker.MagicMock()
        bad_auth_response.status_code = 401
        bad_auth_response.raise_for_status.side_effect = HTTPError(response=bad_auth_response)

        def get_org_repos(method, url, **kwargs):
            if mock_session.headers.get("Authorization") == "token mytoken":
                return bad_auth_response
            elif kwargs["params"]["page"] == 1:
                return response1
            else:
                # End.
                return response2

        mock_session.request.side_effect = get_org_repos
        actual = github_client.available_plugins

        # Still works, even with bad auth.
        assert actual == {"ape_myplugin"}

        # Show we got our log message.
        expected = (
            "Requests are not authorized! GITHUB_ACCESS_TOKEN is likely "
            "expired; received 401 when attempted to use it. If you need "
            "GitHub authorization, try resetting your token."
        )
        assert ape_caplog.head == expected



from pathlib import Path

import click

from ape.cli import ape_cli_context

BASE_PATH = Path(__file__).parent.parent
SOURCES_PATH = BASE_PATH / "sources"
ARTIFACTS_PATH = BASE_PATH / "contracts" / "ethereum" / "local"


def _contract_callback(ctx, param, val):
    for ext in ("sol", "vy"):
        path = SOURCES_PATH / f"{val}.{ext}"
        if path.is_file():
            return path

    raise click.BadArgumentUsage(f"Contract not found: {val}")


@click.command()
@ape_cli_context()
@click.argument("contract_path", callback=_contract_callback)
def cli(cli_ctx, contract_path):
    cm = cli_ctx.compiler_manager
    compiler = "vyper" if contract_path.suffix == ".vy" else "solidity"
    code = contract_path.read_text(encoding="utf-8")
    destination = ARTIFACTS_PATH / f"{contract_path.stem}.json"
    contract_type = cm.compile_source(compiler, code, contractName=contract_path.stem).contract_type

    if contract_type.source_id is None:
        contract_type.source_id = f"{contract_path.stem}.json"

    json_text = contract_type.model_dump_json()
    destination.unlink()
    destination.write_text(json_text, encoding="utf-8")
    click.echo("Done!")



from collections import defaultdict
from collections.abc import Iterator
from contextlib import contextmanager
from datetime import datetime
from functools import cached_property, partial, singledispatchmethod
from statistics import mean, median
from typing import IO, TYPE_CHECKING, cast

import pandas as pd
from rich.box import SIMPLE
from rich.table import Table

from ape.api.address import Address, BaseAddress
from ape.api.providers import BlockAPI
from ape.api.query import (
    AccountTransactionQuery,
    BlockQuery,
    extract_fields,
    validate_and_expand_columns,
)
from ape.api.transactions import ReceiptAPI
from ape.exceptions import (
    APINotImplementedError,
    BlockNotFoundError,
    ChainError,
    ContractNotFoundError,
    ProviderNotConnectedError,
    QueryEngineError,
    TransactionNotFoundError,
    UnknownSnapshotError,
)
from ape.logging import get_rich_console, logger
from ape.managers._contractscache import ContractCache
from ape.managers.base import BaseManager
from ape.types.address import AddressType
from ape.utils.basemodel import BaseInterfaceModel
from ape.utils.misc import ZERO_ADDRESS, is_evm_precompile, is_zero_hex, log_instead_of_fail

if TYPE_CHECKING:
    from rich.console import Console as RichConsole

    from ape.types import BlockID, ContractCode, GasReport, SnapshotID, SourceTraceback


class BlockContainer(BaseManager):
    """
    A list of blocks on the chain.

    Usages example::

        from ape import chain

        latest_block = chain.blocks[-1]

    """

    @property
    def head(self) -> BlockAPI:
        """
        The latest block.
        """
        return self.provider.get_block("latest")

    @property
    def height(self) -> int:
        """
        The latest block number.
        """
        try:
            head = self.head
        except BlockNotFoundError:
            return 0

        if head.number is None:
            raise ChainError("Latest block has no number.")

        return head.number

    @property
    def network_confirmations(self) -> int:
        return self.provider.network.required_confirmations

    def __getitem__(self, block_number: int) -> BlockAPI:
        """
        Get a block by number. Negative numbers start at the chain head and
        move backwards. For example, ``-1`` would be the latest block and
        ``-2`` would be the block prior to that one, and so on.

        Args:
            block_number (int): The number of the block to get.

        Returns:
            :class:`~ape.api.providers.BlockAPI`
        """

        if block_number < 0:
            block_number = len(self) + block_number

        return self.provider.get_block(block_number)

    def __len__(self) -> int:
        """
        The number of blocks in the chain.

        Returns:
            int
        """

        return self.height + 1

    def __iter__(self) -> Iterator[BlockAPI]:
        """
        Iterate over all the current blocks.

        Returns:
            Iterator[:class:`~ape.api.providers.BlockAPI`]
        """

        return self.range(len(self))

    def query(
        self,
        *columns: str,
        start_block: int = 0,
        stop_block: int | None = None,
        step: int = 1,
        engine_to_use: str | None = None,
    ) -> pd.DataFrame:
        """
        A method for querying blocks and returning an Iterator. If you
        do not provide a starting block, the 0 block is assumed. If you do not
        provide a stopping block, the last block is assumed. You can pass
        ``engine_to_use`` to short-circuit engine selection.

        Raises:
            :class:`~ape.exceptions.ChainError`: When ``stop_block`` is greater
              than the chain length.

        Args:
            *columns (str): columns in the DataFrame to return
            start_block (int): The first block, by number, to include in the
              query. Defaults to 0.
            stop_block (int | None): The last block, by number, to include
              in the query. Defaults to the latest block.
            step (int): The number of blocks to iterate between block numbers.
              Defaults to ``1``.
            engine_to_use (str | None): query engine to use, bypasses query
              engine selection algorithm.

        Returns:
            pd.DataFrame
        """

        if start_block < 0:
            start_block = len(self) + start_block

        if stop_block is None:
            stop_block = self.height

        elif stop_block < 0:
            stop_block = len(self) + stop_block

        elif stop_block > len(self):
            raise ChainError(
                f"'stop={stop_block}' cannot be greater than the chain length ({self.height})."
            )

        query = BlockQuery(
            columns=list(columns),
            start_block=start_block,
            stop_block=stop_block,
            step=step,
        )

        blocks = self.query_manager.query(query, engine_to_use=engine_to_use)
        columns: list[str] = validate_and_expand_columns(  # type: ignore
            columns, self.head.__class__
        )
        extraction = partial(extract_fields, columns=columns)
        data = map(lambda b: extraction(b), blocks)
        return pd.DataFrame(columns=columns, data=data)

    def range(
        self,
        start_or_stop: int,
        stop: int | None = None,
        step: int = 1,
        engine_to_use: str | None = None,
    ) -> Iterator[BlockAPI]:
        """
        Iterate over blocks. Works similarly to python ``range()``.

        Raises:
            :class:`~ape.exceptions.ChainError`: When ``stop`` is greater
                than the chain length.
            :class:`~ape.exceptions.ChainError`: When ``stop`` is less
                than ``start_block``.
            :class:`~ape.exceptions.ChainError`: When ``stop`` is less
                than 0.
            :class:`~ape.exceptions.ChainError`: When ``start`` is less
                than 0.

        Args:
            start_or_stop (int): When given just a single value, it is the stop.
              Otherwise, it is the start. This mimics the behavior of ``range``
              built-in Python function.
            stop (int | None): The block number to stop before. Also the total
              number of blocks to get. If not setting a start value, is set by
              the first argument.
            step (int | None): The value to increment by. Defaults to ``1``.
             number of blocks to get. Defaults to the latest block.
            engine_to_use (str | None): query engine to use, bypasses query
              engine selection algorithm.

        Returns:
            Iterator[:class:`~ape.api.providers.BlockAPI`]
        """

        if stop is None:
            stop = start_or_stop
            start = 0
        else:
            start = start_or_stop

        if stop > len(self):
            raise ChainError(
                f"'stop={stop}' cannot be greater than the chain length ({len(self)}). "
                f"Use '{self.poll_blocks.__name__}()' to wait for future blocks."
            )

        # Note: the range `stop_block` is a non-inclusive stop, while the
        #       `.query` method uses an inclusive stop, so we must adjust downwards.
        query = BlockQuery(
            columns=list(self.head.model_fields),  # TODO: fetch the block fields from EcosystemAPI
            start_block=start,
            stop_block=stop - 1,
            step=step,
        )

        blocks = self.query_manager.query(query, engine_to_use=engine_to_use)
        yield from cast(Iterator[BlockAPI], blocks)

    def poll_blocks(
        self,
        start_block: int | None = None,
        stop_block: int | None = None,
        required_confirmations: int | None = None,
        new_block_timeout: int | None = None,
    ) -> Iterator[BlockAPI]:
        """
        Poll new blocks. Optionally set a start block to include historical blocks.

        **NOTE**: When a chain reorganization occurs, this method logs an error and
        yields the missed blocks, even if they were previously yielded with different
        block numbers.

        **NOTE**: This is a daemon method; it does not terminate unless an exception occurs
        or a ``stop_block`` is given.

        Usage example::

            from ape import chain

            for new_block in chain.blocks.poll_blocks():
                print(f"New block found: number={new_block.number}")

        Args:
            start_block (int | None): The block number to start with. Defaults to the pending
              block number.
            stop_block (int | None): Optionally set a future block number to stop at.
              Defaults to never-ending.
            required_confirmations (int | None): The amount of confirmations to wait
              before yielding the block. The more confirmations, the less likely a reorg will occur.
              Defaults to the network's configured required confirmations.
            new_block_timeout (float | None): The amount of time to wait for a new block before
              timing out. Defaults to 10 seconds for local networks or ``50 * block_time`` for live
              networks.

        Returns:
            Iterator[:class:`~ape.api.providers.BlockAPI`]
        """

        if required_confirmations is None:
            required_confirmations = self.network_confirmations

        if stop_block is not None and stop_block <= self.chain_manager.blocks.height:
            raise ValueError("'stop' argument must be in the future.")

        # Get number of last block with the necessary amount of confirmations.
        block = None

        head_minus_confirms = self.height - required_confirmations
        if start_block is not None and start_block <= head_minus_confirms:
            # Front-load historical blocks.
            for block in self.range(start_block, head_minus_confirms + 1):
                yield block

        yield from self.provider.poll_blocks(
            stop_block=stop_block,
            required_confirmations=required_confirmations,
            new_block_timeout=new_block_timeout,
        )


class AccountHistory(BaseInterfaceModel):
    """
    A container mapping account addresses to the transaction from the active session.
    """

    address: AddressType
    """
    The address to get history for.
    """

    sessional: list[ReceiptAPI] = []
    """
    The receipts from the current Python session.
    """

    @property
    def outgoing(self) -> Iterator[ReceiptAPI]:
        """
        All outgoing transactions, from earliest to latest.
        """

        start_nonce = 0
        stop_nonce = len(self) - 1  # just to cache this value

        # TODO: Add ephemeral network sessional history to `ape-cache` instead,
        #       and remove this (replace with `yield from iter(self[:len(self)])`)
        for receipt in self.sessional:
            if receipt.nonce is None:
                # Not an on-chain receipt? idk - has only seen as anomaly in tests.
                continue

            elif receipt.nonce < start_nonce:
                raise QueryEngineError("Sessional history corrupted")

            elif receipt.nonce > start_nonce:
                # NOTE: There's a gap in our sessional history, so fetch from query engine
                yield from iter(self[start_nonce : receipt.nonce + 1])

            yield receipt
            start_nonce = receipt.nonce + 1  # start next loop on the next item

        if start_nonce != stop_nonce:
            # NOTE: there is no more session history, so just return query engine iterator
            yield from iter(self[start_nonce : stop_nonce + 1])

    def query(
        self,
        *columns: str,
        start_nonce: int = 0,
        stop_nonce: int | None = None,
        engine_to_use: str | None = None,
    ) -> pd.DataFrame:
        """
        A method for querying transactions made by an account and returning an Iterator.
        If you do not provide a starting nonce, the first transaction is assumed.
        If you do not provide a stopping block, the last transaction is assumed.
        You can pass ``engine_to_use`` to short-circuit engine selection.

        Raises:
            :class:`~ape.exceptions.ChainError`: When ``stop_nonce`` is greater
              than the account's current nonce.

        Args:
            *columns (str): columns in the DataFrame to return
            start_nonce (int): The first transaction, by nonce, to include in the
              query. Defaults to 0.
            stop_nonce (int | None): The last transaction, by nonce, to include
              in the query. Defaults to the latest transaction.
            engine_to_use (str | None): query engine to use, bypasses query
              engine selection algorithm.

        Returns:
            pd.DataFrame
        """

        if start_nonce < 0:
            start_nonce = len(self) + start_nonce

        if stop_nonce is None:
            stop_nonce = len(self)

        elif stop_nonce < 0:
            stop_nonce = len(self) + stop_nonce

        elif stop_nonce > len(self):
            raise ChainError(
                f"'stop={stop_nonce}' cannot be greater than account's current nonce ({len(self)})."
            )

        query = AccountTransactionQuery(
            columns=list(columns),
            account=self.address,
            start_nonce=start_nonce,
            stop_nonce=stop_nonce,
        )

        txns = self.query_manager.query(query, engine_to_use=engine_to_use)
        columns = validate_and_expand_columns(columns, ReceiptAPI)  # type: ignore
        extraction = partial(extract_fields, columns=columns)
        data = map(lambda tx: extraction(tx), txns)
        return pd.DataFrame(columns=columns, data=data)

    def __iter__(self) -> Iterator[ReceiptAPI]:  # type: ignore[override]
        yield from self.outgoing

    def __len__(self) -> int:
        """
        The transaction count of the address.
        """

        return self.provider.get_nonce(self.address)

    @singledispatchmethod
    def __getitem__(self, index):
        raise IndexError(f"Can't handle type {type(index)}")

    @__getitem__.register
    def __getitem_int(self, index: int) -> ReceiptAPI:
        if index < 0:
            index += len(self)

        try:
            return cast(
                ReceiptAPI,
                next(
                    self.query_manager.query(
                        AccountTransactionQuery(
                            columns=list(ReceiptAPI.__pydantic_fields__),
                            account=self.address,
                            start_nonce=index,
                            stop_nonce=index,
                        )
                    )
                ),
            )
        except StopIteration as e:
            raise IndexError(f"index {index} out of range") from e

    @__getitem__.register
    def __getitem_slice(self, indices: slice) -> list[ReceiptAPI]:
        start, stop, step = (
            indices.start or 0,
            indices.stop or len(self),
            indices.step or 1,
        )

        if start < 0:
            start += len(self)

        if stop < 0:
            stop += len(self)

        elif stop > len(self):
            raise ChainError(
                f"'stop={stop}' cannot be greater than account's current nonce ({len(self)})."
            )

        if stop <= start:
            return []  # nothing to query

        return cast(
            list[ReceiptAPI],
            list(
                self.query_manager.query(
                    AccountTransactionQuery(
                        columns=list(ReceiptAPI.__pydantic_fields__),
                        account=self.address,
                        start_nonce=start,
                        stop_nonce=stop - 1,
                        step=step,
                    )
                )
            ),
        )

    def append(self, receipt: ReceiptAPI):
        """
        Add a receipt to the sessional cache.

        Args:
            receipt (:class:`~ape.api.transactions.ReceiptAPI`): The receipt to append.
        """

        if receipt.txn_hash not in [x.txn_hash for x in self.sessional]:
            self.sessional.append(receipt)

    def revert_to_block(self, block_number: int):
        """
        Remove all receipts after the given block number.

        Args:
            block_number (int): The block number to revert to.
        """

        self.sessional = [x for x in self.sessional if x.block_number <= block_number]


class TransactionHistory(BaseManager):
    """
    A container mapping Transaction History to the transaction from the active session.
    """

    _account_history_cache: dict[AddressType, AccountHistory] = {}
    _hash_to_receipt_map: dict[str, ReceiptAPI] = {}

    @singledispatchmethod
    def __getitem__(self, key):
        raise NotImplementedError(f"Cannot use type of {type(key)} as Index")

    @__getitem__.register
    def __getitem_base_address(self, address: BaseAddress) -> AccountHistory:
        return self._get_account_history(address)

    @__getitem__.register
    def __getitem_str(self, account_or_hash: str) -> AccountHistory | ReceiptAPI:
        """
        Get a receipt from the history by its transaction hash.
        If the receipt is not currently cached, will use the provider
        to retrieve it.

        Args:
            account_or_hash (str): The hash of the transaction.

        Returns:
            :class:`~ape.api.transactions.ReceiptAPI`: The receipt.
        """

        def _get_receipt() -> ReceiptAPI | None:
            try:
                return self._get_receipt(account_or_hash)
            except Exception:
                return None

        is_account = False
        if not account_or_hash.startswith("0x"):
            # Attempt converting.
            try:
                account_or_hash = self.conversion_manager.convert(account_or_hash, AddressType)
            except Exception:
                # Pretend this never happened.
                pass
            else:
                is_account = True

        try:
            address = self.provider.network.ecosystem.decode_address(account_or_hash)
            history = self._get_account_history(address)
            if len(history) > 0:
                return history

        except Exception as err:
            msg = f"'{account_or_hash}' is not a known address or transaction hash."
            if is_account:
                raise ChainError(msg) from err

            # Try to treat as transaction hash.
            elif receipt := _get_receipt():
                return receipt

            # Not an account or tx hash (with success).
            raise ChainError(msg) from err

        # No account history found. Check for transaction hash.
        if receipt := _get_receipt():
            return receipt

        # Nothing found. Return empty history
        return history

    def _get_receipt(self, txn_hash: str) -> ReceiptAPI:
        if cached_receipt := self._hash_to_receipt_map.get(txn_hash):
            return cached_receipt

        receipt = self.provider.get_receipt(txn_hash, timeout=0)
        self.append(receipt)
        return receipt

    def append(self, txn_receipt: ReceiptAPI):
        """
        Add a transaction to the cache This is useful for sessional-transactions.

        Raises:
            :class:`~ape.exceptions.ChainError`: When trying to append a transaction
              receipt that is already in the list.

        Args:
            txn_receipt (:class:`~ape.api.transactions.ReceiptAPI`): The transaction receipt.
        """

        self._hash_to_receipt_map[txn_receipt.txn_hash] = txn_receipt
        key = txn_receipt.sender or ZERO_ADDRESS
        address = self.conversion_manager.convert(key, AddressType)
        if address not in self._account_history_cache:
            self._account_history_cache[address] = AccountHistory(address=address)

        self._account_history_cache[address].append(txn_receipt)

    def revert_to_block(self, block_number: int):
        """
        Remove all receipts past the given block number.

        Args:
            block_number (int): The block number to revert to.
        """
        self._hash_to_receipt_map = {
            h: r for h, r in self._hash_to_receipt_map.items() if r.block_number <= block_number
        }
        for account_history in self._account_history_cache.values():
            account_history.revert_to_block(block_number)

    def _get_account_history(self, address: BaseAddress | AddressType) -> AccountHistory:
        address_key: AddressType = self.conversion_manager.convert(address, AddressType)

        if address_key not in self._account_history_cache:
            self._account_history_cache[address_key] = AccountHistory(address=address_key)

        return self._account_history_cache[address_key]


class ReportManager(BaseManager):
    """
    A class representing the active Ape session. Useful for tracking data and
    building reports.

    **NOTE**: This class is not part of the public API.
    """

    def show_gas(self, report: "GasReport", file: IO[str] | None = None):
        tables: list[Table] = []

        for contract_id, method_calls in report.items():
            title = f"{contract_id} Gas"
            table = Table(title=title, box=SIMPLE)
            table.add_column("Method")
            table.add_column("Times called", justify="right")
            table.add_column("Min.", justify="right")
            table.add_column("Max.", justify="right")
            table.add_column("Mean", justify="right")
            table.add_column("Median", justify="right")
            has_at_least_1_row = False

            for method_call, gases in sorted(method_calls.items()):
                if not gases:
                    continue

                if not method_call or is_zero_hex(method_call) or is_evm_precompile(method_call):
                    continue

                elif method_call == "__new__":
                    # Looks better in the gas report.
                    method_call = "__init__"

                has_at_least_1_row = True
                table.add_row(
                    method_call,
                    f"{len(gases)}",
                    f"{min(gases)}",
                    f"{max(gases)}",
                    f"{int(round(mean(gases)))}",
                    f"{int(round(median(gases)))}",
                )

            if has_at_least_1_row:
                tables.append(table)

        self.echo(*tables, file=file)

    def echo(self, *rich_items, file: IO[str] | None = None, console: "RichConsole | None" = None):
        console = console or get_rich_console(file)
        console.print(*rich_items)

    def show_source_traceback(
        self,
        traceback: "SourceTraceback",
        file: IO[str] | None = None,
        console: "RichConsole | None" = None,
        failing: bool = True,
    ):
        console = console or get_rich_console(file)
        style = "red" if failing else None
        console.print(str(traceback), style=style)

    def show_events(
        self, events: list, file: IO[str] | None = None, console: "RichConsole | None" = None
    ):
        console = console or get_rich_console(file)
        console.print("Events emitted:")
        for event in events:
            console.print(event)

    def _get_console(self, *args, **kwargs):
        # TODO: Delete this method in v0.9.
        # It only exists for backwards compat.
        return get_rich_console(*args, **kwargs)


class ChainManager(BaseManager):
    """
    A class for managing the state of the active blockchain.
    Also, handy for querying data about the chain and managing local caches.
    Access the chain manager singleton from the root ``ape`` namespace.

    Usage example::

        from ape import chain
    """

    _snapshots: defaultdict = defaultdict(list)  # chain_id -> snapshots
    _chain_id_map: dict[str, int] = {}
    _block_container_map: dict[int, BlockContainer] = {}
    _transaction_history_map: dict[int, TransactionHistory] = {}
    _reports: ReportManager = ReportManager()
    _code: dict[str, dict[str, dict[AddressType, "ContractCode"]]] = {}

    @cached_property
    def contracts(self) -> ContractCache:
        """
        A manager for cached contract-types, proxy info, and more.
        """
        return ContractCache()

    @property
    def blocks(self) -> BlockContainer:
        """
        The list of blocks on the chain.
        """
        if self.chain_id not in self._block_container_map:
            blocks = BlockContainer()
            self._block_container_map[self.chain_id] = blocks

        return self._block_container_map[self.chain_id]

    @property
    def history(self) -> TransactionHistory:
        """
        A mapping of transactions from the active session to the account responsible.
        """
        try:
            chain_id = self.chain_id
        except ProviderNotConnectedError:
            return TransactionHistory()  # Empty list.

        if chain_id not in self._transaction_history_map:
            history = TransactionHistory()
            self._transaction_history_map[chain_id] = history

        return self._transaction_history_map[chain_id]

    @property
    def chain_id(self) -> int:
        """
        The blockchain ID.
        See `ChainList <https://chainlist.org/>`__ for a comprehensive list of IDs.
        """
        network_name = self.provider.network.name
        if network_name not in self._chain_id_map:
            self._chain_id_map[network_name] = self.provider.chain_id

        return self._chain_id_map[network_name]

    @property
    def gas_price(self) -> int:
        """
        The price for what it costs to transact.
        """
        return self.provider.gas_price

    @property
    def base_fee(self) -> int:
        """
        The minimum value required to get your transaction included on the next block.
        Only providers that implement `EIP-1559 <https://eips.ethereum.org/EIPS/eip-1559>`__
        will use this property.

        Raises:
            NotImplementedError: When this provider does not implement
              `EIP-1559 <https://eips.ethereum.org/EIPS/eip-1559>`__.
        """
        return self.provider.base_fee

    @property
    def pending_timestamp(self) -> int:
        """
        The current epoch time of the chain, as an ``int``.
        You can also set the timestamp for development purposes.

        Usage example::

            from ape import chain

            chain.pending_timestamp += 3600
        """
        return self.provider.get_block("pending").timestamp

    @pending_timestamp.setter
    def pending_timestamp(self, new_value: int | str):
        self.provider.set_timestamp(self.conversion_manager.convert(new_value, int))

    def get_block_at(
        self,
        time: datetime | int,
        block_time_secs: int | None = None,
    ) -> "BlockAPI":
        """
        Search for a block closest to time ``time``.

        ```{important}
        This method returns the closest block at or after ``time``.
        ```

        Args:
            time: (datetime | int): Time of block to search for.
            block_time_secs: (int | None):
                The average amount of time between blocks. Influenced the search algorithm.
                Defaults to the configured block time for this chain.

        Returns:
            (~:class:`BlockAPI`): The first block mined at or after ``time`` .
        """

        if isinstance(time, datetime):
            time = int(time.timestamp())

        if time < (genesis := self.blocks[0]).timestamp:
            return genesis

        elif time > (head := self.blocks.head).timestamp:
            raise ValueError(f"Time {time} is after head.")
        assert head.number is not None  # mypy happy

        # NOTE: This is only an estimate of the block time, used for the algorithm
        block_time_secs = block_time_secs or self.network_manager.network.block_time or 1

        est_blocks = int((self.pending_timestamp - time) / block_time_secs)
        block = self.blocks[max(head.number - est_blocks, 0)]
        assert block.number is not None  # mypy happy

        # NOTE: Ordering here favors having a block with `.timestamp` that is after `time`
        while time > block.timestamp and (
            est_blocks := int((time - block.timestamp) / block_time_secs)
        ):
            block = self.blocks[min(block.number + est_blocks, head.number)]
            assert block.number is not None  # mypy happy

        assert block.number is not None  # mypy happy
        while time < block.timestamp and (
            est_blocks := int((block.timestamp - time) / block_time_secs)
        ):
            block = self.blocks[max(block.number - est_blocks, 0)]
            assert block.number is not None  # mypy happy

        return block

    @log_instead_of_fail(default="<ChainManager>")
    def __repr__(self) -> str:
        props = f"id={self.chain_id}" if self.network_manager.active_provider else "disconnected"
        cls_name = getattr(type(self), "__name__", ChainManager.__name__)
        return f"<{cls_name} ({props})>"

    def snapshot(self) -> "SnapshotID":
        """
        Record the current state of the blockchain with intent to later
        call the method :meth:`~ape.managers.chain.ChainManager.revert`
        to go back to this point. This method is for local networks only.

        Raises:
            NotImplementedError: When the active provider does not support
              snapshotting.

        Returns:
            :class:`~ape.types.SnapshotID`: The snapshot ID.
        """
        chain_id = self.chain_manager.chain_id
        snapshot_id = self.provider.snapshot()
        if snapshot_id not in self._snapshots[chain_id]:
            self._snapshots[chain_id].append(snapshot_id)

        return snapshot_id

    def restore(self, snapshot_id: "SnapshotID | None" = None):
        """
        Regress the current call using the given snapshot ID.
        Allows developers to go back to a previous state.

        Raises:
            NotImplementedError: When the active provider does not support
              snapshotting.
            :class:`~ape.exceptions.UnknownSnapshotError`: When the snapshot ID is not cached.
            :class:`~ape.exceptions.ChainError`: When there are no snapshot IDs to select from.

        Args:
            snapshot_id (:class:`~ape.types.SnapshotID` | None): The snapshot ID. Defaults
              to the most recent snapshot ID.
        """
        chain_id = self.chain_manager.chain_id
        if snapshot_id is None and not self._snapshots[chain_id]:
            raise ChainError("There are no snapshots to revert to.")
        elif snapshot_id is None:
            snapshot_id = self._snapshots[chain_id].pop()
        elif snapshot_id not in self._snapshots[chain_id]:
            raise UnknownSnapshotError(snapshot_id)
        else:
            snapshot_index = self._snapshots[chain_id].index(snapshot_id)
            self._snapshots[chain_id] = self._snapshots[chain_id][:snapshot_index]

        self.provider.restore(snapshot_id)
        self.history.revert_to_block(self.blocks.height)

    @contextmanager
    def isolate(self):
        """
        Run code in an isolated context.
        Requires using a local provider that supports snapshotting.

        Usages example::

            owner = accounts[0]
            with chain.isolate():
                contract = owner.deploy(project.MyContract)
                receipt = contract.fooBar(sender=owner)
        """

        snapshot = None
        try:
            snapshot = self.snapshot()
        except APINotImplementedError:
            logger.warning("Provider does not support snapshotting.")
        pending = self.pending_timestamp

        start_ecosystem_name = self.provider.network.ecosystem.name
        start_network_name = self.provider.network.name
        start_provider_name = self.provider.name

        try:
            yield
        finally:
            if snapshot is None:
                logger.error("Failed to create snapshot.")
                return

            end_ecosystem_name = self.provider.network.ecosystem.name
            end_network_name = self.provider.network.name
            end_provider_name = self.provider.name

            if (
                start_ecosystem_name != end_ecosystem_name
                or start_network_name != end_network_name
                or start_provider_name != end_provider_name
            ):
                logger.warning("Provider changed before isolation completed.")
                return

            self.chain_manager.restore(snapshot)

            try:
                self.pending_timestamp = pending
            except APINotImplementedError:
                # Provider does not support time travel.
                pass

    def mine(
        self,
        num_blocks: int = 1,
        timestamp: int | None = None,
        deltatime: int | None = None,
    ) -> None:
        """
        Mine any given number of blocks.

        Raises:
            ValueError: When a timestamp AND a deltatime argument are both passed

        Args:
            num_blocks (int): Choose the number of blocks to mine.
                Defaults to 1 block.
            timestamp (int | None): Designate a time (in seconds) to begin mining.
                Defaults to None.
            deltatime (int | None): Designate a change in time (in seconds) to begin mining.
                Defaults to None.
        """
        if timestamp and deltatime:
            raise ValueError("Cannot give both `timestamp` and `deltatime` arguments together.")
        if timestamp:
            self.pending_timestamp = timestamp
        elif deltatime:
            self.pending_timestamp += deltatime
        self.provider.mine(num_blocks)

    def get_balance(
        self, address: BaseAddress | AddressType | str, block_id: "BlockID | None" = None
    ) -> int:
        """
        Get the balance of the given address. If ``ape-ens`` is installed,
        you can pass ENS names.

        Args:
            address (BaseAddress, AddressType | str): An address, ENS, or account/contract.
            block_id (:class:`~ape.types.BlockID` | None): The block ID. Defaults to latest.

        Returns:
            int: The account balance.
        """
        if (isinstance(address, str) and not address.startswith("0x")) or not isinstance(
            address, str
        ):
            # Handles accounts, ENS, integers, aliases, everything.
            address = self.account_manager.resolve_address(address)

        return self.provider.get_balance(address, block_id=block_id)

    def set_balance(self, account: BaseAddress | AddressType | str, amount: int | str):
        """
        Set an account balance, only works on development chains.

        Args:
            account (BaseAddress, AddressType | str): The account.
            amount (int | str): The new balance.
        """
        if isinstance(account, BaseAddress):
            account = account.address

        if isinstance(amount, str) and len(str(amount).split(" ")) > 1:
            # Support values like "1000 ETH".
            amount = self.conversion_manager.convert(amount, int)
        elif isinstance(amount, str):
            # Support hex strings
            amount = int(amount, 16)

        return self.provider.set_balance(account, amount)

    def get_receipt(self, transaction_hash: str) -> ReceiptAPI:
        """
        Get a transaction receipt from the chain.

        Args:
            transaction_hash (str): The hash of the transaction.

        Returns:
            :class:`~ape.apt.transactions.ReceiptAPI`
        """
        receipt = self.chain_manager.history[transaction_hash]
        if not isinstance(receipt, ReceiptAPI):
            raise TransactionNotFoundError(transaction_hash=transaction_hash)

        return receipt

    def get_code(self, address: AddressType, block_id: "BlockID | None" = None) -> "ContractCode":
        network = self.provider.network

        # Two reasons to avoid caching:
        #   1. dev networks - chain isolation makes this mess up
        #   2. specifying block_id= kwarg - likely checking if code
        #      exists at the time and shouldn't use cache.
        skip_cache = network.is_dev or block_id is not None
        if skip_cache:
            return self.provider.get_code(address, block_id=block_id)

        self._code.setdefault(network.ecosystem.name, {})
        self._code[network.ecosystem.name].setdefault(network.name, {})
        if address in self._code[network.ecosystem.name][network.name]:
            return self._code[network.ecosystem.name][network.name][address]

        # Get from RPC for the first time AND use cache.
        code = self.provider.get_code(address)
        self._code[network.ecosystem.name][network.name][address] = code
        return code

    def get_delegate(self, address: AddressType) -> BaseAddress | None:
        ecosystem = self.provider.network.ecosystem

        if not (proxy_info := ecosystem.get_proxy_info(address)):
            return None

        # NOTE: Do this every time because it can change?
        self.contracts.cache_proxy_info(address, proxy_info)

        try:
            return self.contracts.instance_at(proxy_info.target)

        except ContractNotFoundError:
            return Address(proxy_info.target)



import pickle
from copy import deepcopy

import pytest
from eth_pydantic_types import HexBytes
from ethpm_types.abi import ABIType, EventABI, EventABIType

from ape.utils.abi import LogInputABICollection, create_struct


@pytest.fixture
def event_abi():
    return EventABI(
        type="event",
        name="NodeOperatorAdded",
        inputs=[
            EventABIType(
                name="nodeOperatorId",
                type="uint256",
                components=None,
                internalType=None,
                indexed=False,
            ),
            EventABIType(
                name="name", type="string", components=None, internalType=None, indexed=False
            ),
            EventABIType(
                name="rewardAddress",
                type="address",
                components=None,
                internalType=None,
                indexed=False,
            ),
            EventABIType(
                name="stakingLimit",
                type="uint64",
                components=None,
                internalType=None,
                indexed=False,
            ),
        ],
        anonymous=False,
    )


@pytest.fixture
def collection(event_abi):
    return LogInputABICollection(event_abi)


@pytest.fixture
def topics():
    return ["0xc52ec0ad7872dae440d886040390c13677df7bf3cca136d8d81e5e5e7dd62ff1"]


@pytest.fixture(scope="module")
def log_data_missing_trailing_zeroes():
    return HexBytes(
        "0x000000000000000000000000000000000000000000000000000000000000001e"
        "000000000000000000000000000000000000000000000000000000000000008000"
        "00000000000000000000005a8b929edbf3ce44526465dd2087ec7efb59a5610000"
        "000000000000000000000000000000000000000000000000000000000000000000"
        "000000000000000000000000000000000000000000000000000000000b4c61756e"
        "63686e6f646573"
    )


def test_decode_data_missing_trailing_zeroes(
    collection, topics, log_data_missing_trailing_zeroes, ape_caplog
):
    """
    This test is for a time where Alchemy gave us log data when it was missing trailing zeroes.
    When using strict=False, it was able to properly decode. In this case, in Ape, we warn
    the user and still proceed to decode the log.
    """
    actual = ape_caplog.assert_last_log_with_retries(
        lambda: collection.decode(topics, log_data_missing_trailing_zeroes),
        "However, we are able to get a value using decode(strict=False)",
    )
    expected = {
        "name": "Launchnodes",
        "nodeOperatorId": 30,
        "rewardAddress": "0x5a8b929edbf3ce44526465dd2087ec7efb59a561",
        "stakingLimit": 0,
    }
    assert actual == expected


def test_decode_topics_missing_leading_zeroes(project):
    # The second value here was the problem before... It has no leading zeroes
    # and eth-abi is very strict about that.
    topics = [
        "0xa84473122c11e32cd505595f246a28418b8ecd6cf819f4e3915363fad1b8f968",
        "0x0141",
        "0x9f3d45ac20ccf04b45028b8080bb191eab93e29f7898ed43acf480dd80bba94d",
    ]

    # NOTE: data isn't really part of the test but still has to be included.
    data = (
        b"\x9c\xe2\xce\xf5\x9b\xf2\xdeu\x83f\xf8s\xdb\x7f&\xef\xab\x9bw\xf7\xcf"
        b"\xe9\xc8I\xb6\xb5@\x04g\xa9)\x86\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00{\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00`\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x07"
        b"Dynamic\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    )
    abi = project.VyperContract.contract_type.events["NumberChange"]
    collection = LogInputABICollection(abi)

    actual = collection.decode(topics, data)
    assert actual["newNum"] == 321  # NOTE: Was a bug where this causes issues.


class TestStruct:
    @pytest.fixture
    def struct(self):
        return create_struct(
            "MyStruct", (ABIType(name="proptest", type="string"),), ("output_value_0",)
        )

    def test_get_and_set_item(self, struct):
        assert struct["proptest"] == "output_value_0"
        struct["proptest"] = "something else"
        assert struct["proptest"] == "something else"

    def test_is_equal(self, struct):
        # Show struct equality works when props are the same.
        assert struct == struct  # even self
        new_struct = deepcopy(struct)
        assert struct == new_struct
        # Show changing a property makes them unequal.
        new_struct["proptest"] = "something else"
        assert struct != new_struct
        # Show testing with other types works w/o erroring.
        assert struct != 47

    def test_contains(self, struct):
        assert "proptest" in struct

    def test_len(self, struct):
        assert len(struct) == 1

    def test_items(self, struct):
        actual = struct.items()
        assert len(actual) == 1
        assert actual[0] == ("proptest", "output_value_0")

    def test_values(self, struct):
        actual = struct.values()
        assert len(actual) == 1
        assert actual[0] == "output_value_0"

    def test_pickle(self, struct):
        actual = pickle.dumps(struct)
        assert isinstance(actual, bytes)

    def test_field_with_same_name_as_method(self):
        struct = create_struct(
            "MyStruct", (ABIType(name="values", type="string"),), ("output_value_0",)
        )
        assert struct.values == "output_value_0"  # Is the field, not the method.



import pytest


@pytest.fixture(scope="session")
def owner(accounts):
    return accounts[0]


@pytest.fixture(scope="session")
def contract(project, owner):
    _contract = project.VyperContract.deploy(sender=owner)

    # Show that contract transactions in fixtures appear in gas report
    _contract.setNumber(999, sender=owner)

    return _contract



from ape.exceptions import ApeException


class MulticallException(ApeException):
    pass


class InvalidOption(MulticallException):
    def __init__(self, option_name: str):
        super().__init__(f"Option '{option_name}' not supported.")


class ValueRequired(MulticallException):
    def __init__(self, amount: int):
        super().__init__(f"This transaction must send at least '{amount / 1e18}' ether.")


class UnsupportedChainError(MulticallException):
    def __init__(self):
        super().__init__("Multicall not supported on this chain.")


class NotExecutedError(MulticallException):
    def __init__(self):
        super().__init__("Multicall not executed yet.")



from collections.abc import Iterator
from typing import TYPE_CHECKING

from .pluggy_patch import PluginType, hookspec

if TYPE_CHECKING:
    from ape.api.explorers import ExplorerAPI
    from ape.api.networks import EcosystemAPI, NetworkAPI
    from ape.api.providers import ProviderAPI


class EcosystemPlugin(PluginType):
    """
    An ecosystem plugin, such as ``ape-ethereum``. See the
    :class:`ape.api.networks.EcosystemAPI` for more information on
    what is required to implement an ecosystem plugin.
    """

    @hookspec
    def ecosystems(self) -> Iterator[type["EcosystemAPI"]]:  # type: ignore[empty-body]
        """
        A hook that must return an iterator of :class:`ape.api.networks.EcosystemAPI`
        subclasses.

        Usage example::

            @plugins.register(plugins.EcosystemPlugin)
            def ecosystems():
                yield Ethereum

        Returns:
            Iterator[type[:class:`~ape.api.networks.EcosystemAPI`]]
        """


class NetworkPlugin(PluginType):
    """
    A network plugin, such as ``mainnet`` or ``ropsten``. Likely, registering networks
    will happen soon after registering the ecosystem, as an ecosystem requires
    networks.
    """

    @hookspec
    def networks(self) -> Iterator[tuple[str, str, type["NetworkAPI"]]]:  # type: ignore[empty-body]
        """
        A hook that must return an iterator of tuples of:

        * the target ecosystem plugin's name
        * the network name
        * a :class:`ape.api.networks.NetworkAPI` subclass

        Usage example::

            @plugins.register(plugins.NetworkPlugin)
            def networks():
                yield "ethereum", "ShibaChain", ShibaNetwork

        Returns:
            Iterator[tuple[str, str, type[:class:`~ape.api.networks.NetworkAPI`]]]
        """


class ProviderPlugin(PluginType):
    """
    A plugin representing a network provider, which is the main API responsible
    for making requests against a blockchain. Example provider plugins projects
    include `ape-infura <https://github.com/ApeWorX/ape-infura>`__ as well as
    `ape-alchemy <https://github.com/ApeWorX/ape-alchemy>`__.
    """

    @hookspec
    def providers(  # type: ignore[empty-body]
        self,
    ) -> Iterator[tuple[str, str, type["ProviderAPI"]]]:
        """
        A hook that must return an iterator of tuples of:

        * the target ecosystem plugin's name
        * the network it works with (which must be valid network in the ecosystem)
        * a :class:`ape.api.providers.ProviderAPI` subclass

        Usage example::

            @plugins.register(plugins.ProviderPlugin)
            def providers():
                yield "ethereum", "local", MyProvider

        Returns:
            Iterator[tuple[str, str, type[:class:`~ape.api.providers.ProviderAPI`]]]
        """


class ExplorerPlugin(PluginType):
    """
    A plugin for a blockchain explorer, such as
    `ape-etherscan <https://github.com/ApeWorX/ape-etherscan>`__.
    """

    @hookspec
    def explorers(  # type: ignore[empty-body]
        self,
    ) -> Iterator[tuple[str, str, type["ExplorerAPI"]]]:
        """
        A hook that must return an iterator of tuples of:

        * the target ecosystem plugin's name
        * the network it works with (which must be valid network in the ecosystem)
        * a :class:`~ape.api.explorers.ExplorerAPI` subclass

        Usage example::

            @plugins.register(plugins.ExplorerPlugin)
            def explorers():
                yield "ethereum", "mainnet", MyBlockExplorer

        Returns:
            Iterator[tuple[str, str, type[:class:`ape.api.explorers.ExplorerAPI`]]]
        """



import re

import pytest

from tests.integration.cli.utils import run_once


@run_once
@pytest.mark.parametrize(
    "args",
    [
        (),
        ("--version",),
        ("--config",),
    ],
)
def test_invocation(ape_cli, runner, args):
    result = runner.invoke(ape_cli, args)
    assert result.exit_code == 0, result.output


@run_once
def test_help(ape_cli, runner):
    result = runner.invoke(ape_cli, "--help")
    assert result.exit_code == 0, result.output
    anything = r"[.\n\s\w`/\-,\)\(:\]\[']*"
    expected = (
        rf"{anything}Core Commands:\n  accounts  "
        rf"Manage local accounts{anything}  "
        rf"test\s*Launches pytest{anything}"
    )
    assert re.match(expected.strip(), result.output)



from pathlib import Path

import pytest

from ape.exceptions import ConfigError
from ape.pytest.config import ConfigWrapper
from ape.pytest.fixtures import FixtureManager, FixtureMap, IsolationManager, SnapshotRegistry
from ape.pytest.runners import PytestApeRunner
from ape.pytest.utils import Scope
from ape.pytest.warnings import InvalidIsolationWarning
from ape_test import ApeTestConfig
from ape_test._watch import run_with_observer
from ape_test.config import IsolationConfig


@pytest.fixture
def create_fixture_info(mocker):
    def fn(name="my_fixture", scope=Scope.FUNCTION.value, params=None, cached_result=None):
        info = mocker.MagicMock()
        info.argname = name
        info.scope = scope
        info.params = params
        info.cached_result = cached_result
        return info

    return fn


@pytest.fixture
def item(mocker, create_fixture_info):
    # foo, bar, and baz are fixtures; param0 and param1 are test-params.
    fixturenames = ["foo", "bar", "baz", "param0", "param1"]
    mock = mocker.MagicMock()
    mock.nodeid = "test_nodeid"
    mock.fixturenames = fixturenames
    mock.session._fixturemanager._arg2fixturedefs = {
        "_session_isolation": [create_fixture_info("_session_isolation", Scope.SESSION.value)],
        "_package_isolation": [create_fixture_info("_package_isolation", Scope.PACKAGE.value)],
        "foo": [create_fixture_info("foo", Scope.SESSION.value, [1, 2, 3])],
        "_module_isolation": [create_fixture_info("_module_isolation", Scope.MODULE.value)],
        "bar": [create_fixture_info("bar", Scope.MODULE.value)],
        "_class_isolation": [create_fixture_info("_class_isolation", Scope.CLASS.value)],
        "baz": [create_fixture_info("baz", Scope.CLASS.value)],
        "_function_isolation": [create_fixture_info("_function_isolation", Scope.FUNCTION.value)],
    }
    return mock


@pytest.fixture
def fixture_map(item):
    return FixtureMap.from_test_item(item)


class TestApeTestConfig:
    def test_balance_set_from_currency_str(self):
        curr_val = "10 Eth"
        data = {"balance": curr_val}
        cfg = ApeTestConfig.model_validate(data)
        actual = cfg.balance
        expected = 10_000_000_000_000_000_000  # 10 ETH in WEI
        assert actual == expected


class TestConfigWrapper:
    @pytest.mark.parametrize(
        "cli_value,config_value", [(True, False), (False, True), (False, False)]
    )
    def test_show_internal(self, mocker, cli_value, config_value):
        pytest_cfg = mocker.MagicMock()
        ape_test_cfg = mocker.MagicMock()
        wrapper = ConfigWrapper(pytest_cfg)
        wrapper.__dict__["ape_test_config"] = ape_test_cfg
        expected = cli_value or config_value  # True if there is a True
        pytest_cfg.getoption.return_value = cli_value
        ape_test_cfg.show_internal = config_value
        assert wrapper.show_internal is expected

    def test_verbosity(self, mocker):
        """
        Show it returns the same as pytest_config's.
        """
        pytest_cfg = mocker.MagicMock()
        pytest_cfg.option.verbose = False
        wrapper = ConfigWrapper(pytest_cfg)
        assert wrapper.verbosity is False

    def test_verbosity_when_no_capture(self, mocker):
        """
        Shows we enable verbose output when no-capture is set.
        """

        def get_opt(name: str):
            return "no" if name == "capture" else None

        pytest_cfg = mocker.MagicMock()
        pytest_cfg.option.verbose = False  # Start off as False
        pytest_cfg.getoption.side_effect = get_opt

        wrapper = ConfigWrapper(pytest_cfg)
        assert wrapper.verbosity is True

    @pytest.mark.parametrize("flag", (True, None))
    def test_isolation_command_line(self, mocker, flag):
        pytest_cfg = mocker.MagicMock()

        def get_opt(name: str):
            if name == "disable_isolation":
                return flag

        pytest_cfg.getoption.side_effect = get_opt
        wrapper = ConfigWrapper(pytest_cfg)

        if flag:
            assert not wrapper.isolation
            for scope in Scope:
                assert not wrapper.get_isolation(scope)

        else:
            assert wrapper.isolation

    def test_isolation_config(self, mocker):
        pytest_cfg = mocker.MagicMock()
        pytest_cfg.getoption.return_value = None
        ape_test_cfg = ApeTestConfig()
        wrapper = ConfigWrapper(pytest_cfg)
        wrapper.__dict__["ape_test_config"] = ape_test_cfg

        # Show can configure isolation as True.
        ape_test_cfg.isolation = True
        for scope in Scope:
            assert wrapper.get_isolation(scope)

        # Show can configure isolation as False.
        ape_test_cfg.isolation = False
        for scope in Scope:
            assert not wrapper.get_isolation(scope)

        # Show can configure individual scopes.
        ape_test_cfg.isolation = IsolationConfig(
            enable_session=True,
            enable_package=False,
            enable_function=True,
        )
        assert wrapper.get_isolation(Scope.SESSION)
        assert not wrapper.get_isolation(Scope.PACKAGE)  # default
        assert wrapper.get_isolation(Scope.MODULE)  # default
        assert wrapper.get_isolation(Scope.FUNCTION)


def test_connect_to_mainnet_by_default(mocker):
    """
    Tests the condition where mainnet is configured as the default network
    and no --network option is passed. It should avoid running the tests
    to be safe.
    """

    cfg = mocker.MagicMock()
    cfg.network = "ethereum:mainnet:node"
    runner = PytestApeRunner(
        cfg, mocker.MagicMock(), mocker.MagicMock(), mocker.MagicMock(), mocker.MagicMock()
    )
    expected = (
        "Default network is mainnet; unable to run tests on mainnet. "
        "Please specify the network using the `--network` flag or "
        "configure a different default network."
    )
    with pytest.raises(ConfigError, match=expected):
        runner._connect()


class TestFixtureManager:
    @pytest.fixture
    def fixture_manager(self, mocker):
        config = mocker.MagicMock()
        isolation = mocker.MagicMock()
        return FixtureManager(config, isolation)

    def test_ape_fixtures(self, fixture_manager):
        actual = fixture_manager._ape_fixtures
        assert "accounts" in actual
        assert "project" in actual
        assert "networks" in actual
        assert "chain" in actual

    def test_is_isolation(self, fixture_manager):
        assert fixture_manager.is_isolation("_session_isolation")
        assert fixture_manager.is_isolation("_package_isolation")
        assert fixture_manager.is_isolation("_module_isolation")
        assert fixture_manager.is_isolation("_class_isolation")
        assert fixture_manager.is_isolation("_function_isolation")
        assert not fixture_manager.is_isolation("my_custom_isolation")

    def tes_is_ape(self, fixture_manager):
        assert fixture_manager.is_ape("accounts")
        assert not fixture_manager.is_ape("acct")

    def test_is_custom(self, fixture_manager):
        assert fixture_manager.is_custom("acct")
        assert not fixture_manager.is_custom("accounts")
        assert not fixture_manager.is_custom("_module_isolation")

    def test_get_fixtures(self, fixture_manager, item):
        with pytest.raises(KeyError):
            assert fixture_manager.get_fixtures(item.nodeid) is None

        actual = fixture_manager.get_fixtures(item)
        assert isinstance(actual, FixtureMap)
        # Now, can use nodeid.
        assert fixture_manager.get_fixtures(item.nodeid) == actual

    def test_is_stateful(self, fixture_manager, item):
        fixture_manager.cache_fixtures(item)
        assert fixture_manager.is_stateful("foo") is None  # Unknown.
        fixture_manager.add_fixture_info("foo", setup_block=1, teardown_block=1)
        assert fixture_manager.is_stateful("foo") is False
        fixture_manager.add_fixture_info("bar", setup_block=1, teardown_block=2)
        assert fixture_manager.is_stateful("bar") is True

    def test_is_stateful_when_not_using_test_provider(self, networks, fixture_manager, item):
        fixture_manager._stateful_fixtures_cache = {}
        live_provider = networks.ethereum.sepolia.get_provider("node")
        provider = networks.active_provider
        networks.active_provider = live_provider
        for fixture in ("foo", "bar"):
            actual = fixture_manager.is_stateful(fixture)
            networks.active_provider = provider
            assert actual is None

    def test_needs_rebase(self, mocker, fixture_manager):
        new_fixtures = ["new_fixture_1"]
        snapshot = mocker.MagicMock()
        actual = fixture_manager.needs_rebase(new_fixtures, snapshot)
        assert actual

        # Show you can disable fixture-rebasing.
        original_value = fixture_manager.config_wrapper.__dict__.pop(
            "enable_fixture_rebasing", True
        )
        fixture_manager.config_wrapper.__dict__["enable_fixture_rebasing"] = False
        actual = fixture_manager.needs_rebase(new_fixtures, snapshot)

        # (put back)
        fixture_manager.config_wrapper.__dict__["enable_fixture_rebasing"] = original_value

        assert not actual

    def test_rebase(self, mocker, fixture_manager, fixture_map, create_fixture_info):
        # We must have already started our module-scope isolation.
        isolation_manager = IsolationManager(fixture_manager.config_wrapper, mocker.MagicMock())
        isolation_manager.snapshots[Scope.MODULE].identifier = "123"
        isolation_manager.snapshots[Scope.MODULE].fixtures = ["bar"]
        fixture_manager.isolation_manager = isolation_manager

        # New session fixture arrives, triggering a rebase.
        fixture_map[Scope.SESSION].append("new_session_fixture")
        fixture_map._item.fixturenames.append("new_session_fixture")
        fixture_map._item.session._fixturemanager._arg2fixturedefs["new_session_fixture"] = [
            create_fixture_info("new_session_fixture", Scope.SESSION)
        ]

        # Cache a module result so we can prove it gets cleared.
        fixture_map._item.session._fixturemanager._arg2fixturedefs["bar"][0].cached_result = "CACHE"

        # Show the module-isolation was cache and gets cleared as well.
        # NOTE: Pytest caches yield-based fixtures as a tuple, even when yields None.
        fixture_map._item.session._fixturemanager._arg2fixturedefs["_module_isolation"][
            0
        ].cached_result = (None, None, None)

        expected = (
            r"Invalid isolation; Ensure session|package|module|class scoped "
            r"fixtures run earlier\. Rebasing fixtures is costly\."
        )
        with pytest.warns(InvalidIsolationWarning, match=expected):
            fixture_manager.rebase(Scope.SESSION, fixture_map)

        # Show that module-level fixtures are invalidated, including the isolation fixture.
        for module_fixture_name in ("bar", "_module_isolation"):
            assert (
                fixture_map._item.session._fixturemanager._arg2fixturedefs[module_fixture_name][
                    0
                ].cached_result
                is None
            )
        # Show that we have reverted our module-level snapshot.
        assert isolation_manager.snapshots[Scope.MODULE].identifier is None


class TestFixtureMap:
    def test_from_test_item(self, item):
        actual = FixtureMap.from_test_item(item)
        assert actual[Scope.SESSION] == ["foo"]
        assert actual[Scope.MODULE] == ["bar"]
        assert actual[Scope.CLASS] == ["baz"]

    def test_names(self, fixture_map):
        """
        Show that we have both the initialized fixtures as well
        as the properly injected isolation fixtures. Order is
        EXTREMELY important here! It determines the order in which
        fixtures run; isolation should run before their sister fixtures.
        Function isolation is expected even when not using other function-scoped
        fixtures. Package isolation is missing because there are no
        package-scoped fixtures being used.
        """
        actual = fixture_map.names
        expected = [
            "_session_isolation",
            "foo",
            "_module_isolation",
            "bar",
            "_class_isolation",
            "baz",
            "_function_isolation",
        ]
        assert actual == expected

    def test_parameters(self, fixture_map):
        actual = fixture_map.parameters
        expected = ["param0", "param1"]
        assert actual == expected

    def test_isolation(self, fixture_map):
        actual = fixture_map.isolation
        expected = [
            "session",
            "module",
            "class",
            "function",
        ]
        assert actual == expected

    def test_parametrized(self, fixture_map):
        actual = fixture_map.parametrized
        assert "foo" in actual
        assert len(actual) == 1

    def test_get_info(self, fixture_map):
        actual = fixture_map.get_info("foo")
        assert len(actual) == 1
        assert actual[0].argname == "foo"
        assert actual[0].scope == Scope.SESSION

    def test_is_known(self, fixture_map):
        assert fixture_map.is_known("foo")
        assert not fixture_map.is_known("param0")

    def test_is_iterating(self, fixture_map):
        assert fixture_map.is_iterating("foo")
        assert not fixture_map.is_iterating("baz")

        # Iterate.
        fixture_map._item.session._fixturemanager._arg2fixturedefs["foo"][0].cached_result = (
            None,
            1,
            None,
        )
        assert fixture_map.is_iterating("foo")

        # Complete.
        fixture_map._item.session._fixturemanager._arg2fixturedefs["foo"][0].cached_result = (
            None,
            3,
            None,
        )
        assert not fixture_map.is_iterating("foo")

    def test_apply_fixturenames(self, fixture_map):
        assert fixture_map._item.fixturenames == ["foo", "bar", "baz", "param0", "param1"]
        fixture_map.apply_fixturenames()
        assert fixture_map._item.fixturenames == [
            "_session_isolation",
            "foo",
            "_module_isolation",
            "bar",
            "_class_isolation",
            "baz",
            "_function_isolation",
            "param0",
            "param1",
        ]


class TestSnapshotRegistry:
    """
    Note: Most isolation-based tests occur in `functional/test_fixtures.py`.
    """

    @pytest.fixture
    def registry(self):
        return SnapshotRegistry()

    def test_get_snapshot_id(self, registry):
        actual = registry.get_snapshot_id(Scope.SESSION)
        assert actual is None

    def test_next_snapshots(self, registry):
        actual = [x for x in registry.next_snapshots(Scope.SESSION)]
        assert actual[0].scope is Scope.PACKAGE
        assert actual[1].scope is Scope.MODULE
        assert actual[2].scope is Scope.CLASS
        assert actual[3].scope is Scope.FUNCTION


class TestIsolationManager:
    """
    Note: Most isolation-based tests occur in `functional/test_fixtures.py`.
    """

    @pytest.fixture
    def isolation_manager(self, mocker):
        config_wrapper = mocker.MagicMock()
        receipt_capture = mocker.MagicMock()
        return IsolationManager(config_wrapper, receipt_capture)

    @pytest.fixture
    def empty_snapshot_registry(self, isolation_manager):
        snapshots = isolation_manager.snapshots
        isolation_manager.snapshots = SnapshotRegistry()
        yield
        isolation_manager.snapshots = snapshots

    def test_get_snapshot(self, isolation_manager):
        actual = isolation_manager.get_snapshot(Scope.SESSION)
        assert actual.scope is Scope.SESSION

    def test_next_snapshots(self, isolation_manager):
        actual = [x for x in isolation_manager.next_snapshots(Scope.SESSION)]
        assert actual[0].scope is Scope.PACKAGE
        assert actual[1].scope is Scope.MODULE
        assert actual[2].scope is Scope.CLASS
        assert actual[3].scope is Scope.FUNCTION

    def test_isolate(
        self, isolation_manager, owner, vyper_contract_instance, empty_snapshot_registry
    ):
        """
        Low-level test simulating how pytest interacts with these yield-based
        isolation fixtures.
        """
        start_number = vyper_contract_instance.myNumber()
        session = isolation_manager.isolation(Scope.SESSION)
        module = isolation_manager.isolation(Scope.MODULE)
        function = isolation_manager.isolation(Scope.FUNCTION)

        expected_session = 10_000_000
        expected_module = 20_000_000
        expected_test = 30_000_000

        # Show we start off clear of snapshots.
        assert all(isolation_manager.snapshots[s].identifier is None for s in Scope), (
            "Setup failed - snapshots not empty"
        )

        # Start session.
        next(session)
        assert isolation_manager.snapshots[Scope.SESSION].identifier is not None
        vyper_contract_instance.setNumber(expected_session, sender=owner)

        # Start module.
        next(module)
        vyper_contract_instance.setNumber(expected_module, sender=owner)

        # Start test.
        next(function)
        vyper_contract_instance.setNumber(expected_test, sender=owner)
        assert vyper_contract_instance.myNumber() == expected_test

        # End test; back to module.
        next(function, None)
        assert vyper_contract_instance.myNumber() == expected_module, "Is not back at module."

        # End module; back to session.
        assert isolation_manager.snapshots[Scope.MODULE].identifier is not None
        next(module, None)
        assert vyper_contract_instance.myNumber() == expected_session, "Is not back at session."

        # Start new module.
        module = isolation_manager.isolation(Scope.MODULE)
        next(module)
        vyper_contract_instance.setNumber(expected_module, sender=owner)

        # Start new test.
        function = isolation_manager.isolation(Scope.FUNCTION)
        next(function)
        vyper_contract_instance.setNumber(expected_test, sender=owner)
        assert vyper_contract_instance.myNumber() == expected_test

        # End test.
        next(function, None)
        assert vyper_contract_instance.myNumber() == expected_module, "(2) Is not back at module."

        # End module.
        assert isolation_manager.snapshots[Scope.MODULE].identifier is not None
        next(module, None)
        assert isolation_manager.snapshots[Scope.MODULE].identifier is None
        assert vyper_contract_instance.myNumber() == expected_session, "(2) Is not back at session."

        # End session.
        next(session, None)
        assert vyper_contract_instance.myNumber() == start_number, "(2) Is not back pre-session."


def test_watch(mocker):
    mock_event_handler = mocker.MagicMock()
    event_handler_patch = mocker.patch("ape_test._watch._create_event_handler")
    event_handler_patch.return_value = mock_event_handler

    mock_observer = mocker.MagicMock()
    observer_patch = mocker.patch("ape_test._watch._create_observer")
    observer_patch.return_value = mock_observer

    run_subprocess_patch = mocker.patch("ape_test._watch.run_subprocess")
    run_main_loop_patch = mocker.patch("ape_test._watch._run_main_loop")
    run_main_loop_patch.side_effect = SystemExit  # Avoid infinite loop.

    # Only passing `-s` so we have an extra arg to test.
    with pytest.raises(SystemExit):
        run_with_observer((Path("contracts"),), 0.1, "-s")

    # The observer started, then the main runner exits, and the observer stops + joins.
    assert mock_observer.start.call_count == 1
    assert mock_observer.stop.call_count == 1
    assert mock_observer.join.call_count == 1

    # NOTE: We had a bug once where the args it received were not strings.
    #   (wasn't deconstructing), so this check is important.
    run_subprocess_patch.assert_called_once_with(["ape", "test", "-s"])



from abc import abstractmethod
from functools import cached_property
from pathlib import Path

from pydantic import Field, field_validator

from ape.api.config import ApeConfig
from ape.utils.basemodel import BaseInterfaceModel


class DependencyAPI(BaseInterfaceModel):
    """
    An API for obtaining sources.
    """

    name: str
    """
    The package-name of the dependency.
    """

    config_override: dict = Field(default_factory=dict, repr=False)
    """
    Set different config than what Ape can deduce.
    """

    @property
    @abstractmethod
    def package_id(self) -> str:
        """
        The full name of the package, used for storage.
        Example: ``OpenZeppelin/openzeppelin-contracts``.
        """

    @property
    @abstractmethod
    def version_id(self) -> str:
        """
        The ID to use as the sub-directory in the download cache.
        Most often, this is either a version number or a branch name.
        """

    @property
    @abstractmethod
    def uri(self) -> str:
        """
        The URI for the package.
        """

    @abstractmethod
    def fetch(self, destination: Path):
        """
        Fetch the dependency. E.g. for GitHub dependency,
        download the files to the destination.

        Args:
            destination (Path): The destination for the dependency
              files.
        """

    @field_validator("name", mode="before")
    @classmethod
    def validate_name(cls, value):
        return (value or "").lower().replace("_", "-")

    def __hash__(self) -> int:
        return hash(f"{self.package_id}@{self.version_id}")


class ProjectAPI(BaseInterfaceModel):
    """
    An API for recognizing different project types,
    such as brownie projects versus ape projects.
    NOTE: This assumed the project sources are available and unpacked.
    Use :class:`~ape.api.projects.DependencyAPI` to fetch different
    projects. The main task of the project API is to generate
    a configuration needed to compile in Ape.
    """

    path: Path
    """
    The location of the project.
    """

    @property
    @abstractmethod
    def is_valid(self) -> bool:
        """
        Return ``True`` when detecting a project of this type.
        """

    @abstractmethod
    def extract_config(self, **overrides) -> "ApeConfig":
        """
        Extra configuration from the project so that
        Ape understands the dependencies and how to compile everything.

        Args:
            **overrides: Config overrides.

        Returns:
            :class:`~ape.managers.config.ApeConfig`
        """

    @classmethod
    def attempt_validate(cls, **kwargs) -> "ProjectAPI | None":
        try:
            instance = cls(**kwargs)
        except ValueError:
            return None

        return instance if instance.is_valid else None


class ApeProject(ProjectAPI):
    """
    The default ProjectAPI implementation.
    """

    CONFIG_FILE_NAME: str = "ape-config"
    EXTENSIONS: tuple[str, ...] = (".yaml", ".yml", ".json")

    @property
    def is_valid(self) -> bool:
        return True  # If all else fails, treat as a default Ape project.

    @cached_property
    def config_file(self) -> Path:
        if self._using_pyproject_toml:
            return self._pyproject_toml

        # else: check for an ape-config file.
        for ext in self.EXTENSIONS:
            path = self.path / f"{self.CONFIG_FILE_NAME}{ext}"
            if path.is_file():
                return path

        # Default: non-existing ape-config.yaml file.
        return self.path / f"{self.CONFIG_FILE_NAME}.yaml"

    @property
    def _pyproject_toml(self) -> Path:
        return self.path / "pyproject.toml"

    @property
    def _using_pyproject_toml(self) -> bool:
        return self._pyproject_toml.is_file() and "[tool.ape" in self._pyproject_toml.read_text()

    def extract_config(self, **overrides) -> ApeConfig:
        return ApeConfig.validate_file(self.config_file, **overrides)



from abc import abstractmethod
from functools import cached_property
from typing import TYPE_CHECKING, Any

from ape.exceptions import AccountsError, ConversionError
from ape.types.address import AddressType
from ape.types.units import CurrencyValue
from ape.utils.basemodel import BaseInterface
from ape.utils.misc import log_instead_of_fail

if TYPE_CHECKING:
    from ape.api.transactions import ReceiptAPI, TransactionAPI
    from ape.managers.chain import AccountHistory
    from ape.types.vm import ContractCode


class BaseAddress(BaseInterface):
    """
    A base address API class. All account-types subclass this type.
    """

    @property
    def _base_dir_values(self) -> list[str]:
        """
        This exists because when you call ``dir(BaseAddress)``, you get the type's return
        value and not the instances. This allows base-classes to make use of shared
        ``IPython`` ``__dir__`` values.
        """

        # NOTE: mypy is confused by properties.
        #  https://github.com/python/typing/issues/1112

        return [
            str(BaseAddress.address.fget.__name__),  # type: ignore[attr-defined]
            str(BaseAddress.balance.fget.__name__),  # type: ignore[attr-defined]
            str(BaseAddress.code.fget.__name__),  # type: ignore[attr-defined]
            str(BaseAddress.codesize.fget.__name__),  # type: ignore[attr-defined]
            str(BaseAddress.nonce.fget.__name__),  # type: ignore[attr-defined]
            str(BaseAddress.is_contract.fget.__name__),  # type: ignore[attr-defined]
            "provider",  # Is a class property
        ]

    @property
    @abstractmethod
    def address(self) -> AddressType:
        """
        The address of this account. Subclasses must override and provide this value.
        """

    def __eq__(self, other: object) -> bool:
        """
        Compares :class:`~ape.api.BaseAddress` or ``str`` objects by converting to
        :class:`~ape.types.address.AddressType`.

        Returns:
            bool: comparison result
        """

        convert = self.conversion_manager.convert

        try:
            return convert(self, AddressType) == convert(other, AddressType)
        except ConversionError:
            # Check other __eq__
            return NotImplemented

    def __dir__(self) -> list[str]:
        """
        Display methods to IPython on ``a.[TAB]`` tab completion.
        Overridden to lessen amount of methods shown to only those that are useful.

        Returns:
            list[str]: Method names that IPython uses for tab completion.
        """
        return self._base_dir_values

    @log_instead_of_fail(default="<BaseAddress>")
    def __repr__(self) -> str:
        cls_name = getattr(type(self), "__name__", BaseAddress.__name__)
        return f"<{cls_name} {self.address}>"

    def __str__(self) -> str:
        """
        Convert this class to a ``str`` address.

        Returns:
            str: The stringified address.
        """
        return self.address

    def __hash__(self) -> int:
        """Return consistent hash for all address representations with the same value."""
        return hash(int(self.address, base=16))

    def __call__(self, **kwargs) -> "ReceiptAPI":
        """
        Call this address directly. For contracts, this may mean invoking their
        default handler.

        Args:
            **kwargs: Transaction arguments, such as ``sender`` or ``data``.

        Returns:
            :class:`~ape.api.transactions.ReceiptAPI`
        """

        txn = self.as_transaction(**kwargs)
        if "sender" in kwargs:
            if hasattr(kwargs["sender"], "call"):
                # AccountAPI
                sender = kwargs["sender"]
                return sender.call(txn, **kwargs)

            elif hasattr(kwargs["sender"], "prepare_transaction"):
                # BaseAddress (likely, a ContractInstance)
                prepare_transaction = kwargs["sender"].prepare_transaction(txn)
                return self.provider.send_transaction(prepare_transaction)

        elif "sender" not in kwargs and self.account_manager.default_sender is not None:
            return self.account_manager.default_sender.call(txn, **kwargs)

        return self.provider.send_transaction(txn)

    @property
    def nonce(self) -> int:
        """
        The number of transactions associated with the address.
        """

        return self.provider.get_nonce(self.address)

    @property
    def balance(self) -> int:
        """
        The total balance of the account.
        """
        bal = self.provider.get_balance(self.address)
        # By using CurrencyValue, we can compare with
        # strings like "1 ether".
        return CurrencyValue(bal)

    # @balance.setter
    # NOTE: commented out because of failure noted within `__setattr__`
    def _set_balance_(self, value: Any):
        if isinstance(value, str):
            value = self.conversion_manager.convert(value, int)

        self.provider.set_balance(self.address, value)

    def __setattr__(self, attr: str, value: Any) -> None:
        # NOTE: Need to do this until https://github.com/pydantic/pydantic/pull/2625 is figured out
        if attr == "balance":
            self._set_balance_(value)

        else:
            super().__setattr__(attr, value)

    @property
    def code(self) -> "ContractCode":
        """
        The raw bytes of the smart-contract code at the address.
        """
        # NOTE: Chain manager handles code caching.
        return self.chain_manager.get_code(self.address)

    @property
    def codesize(self) -> int:
        """
        The number of bytes in the smart contract.
        """
        code = self.code
        return len(code) if isinstance(code, bytes) else len(bytes.fromhex(code.lstrip("0x")))

    @property
    def is_contract(self) -> bool:
        """
        ``True`` when there is code associated with the address.
        """
        return self.codesize > 0

    @property
    def delegate(self) -> "BaseAddress | None":
        """
        Check and see if Account has a "delegate" contract, which is a contract that this account
        delegates functionality to. This could be from many contexts, such as a Smart Wallet like
        Safe (https://github.com/ApeWorX/ape-safe) which has a Singleton class it forwards to, or
        an EOA using an EIP7702-style delegate. Returning ``None`` means that the account does not
        have a delegate.

        The default behavior is to use `:class:~ape.managers.ChainManager.get_delegate` to check if
        the account has a proxy, such as ``SafeProxy`` for ``ape-safe`` or an EIP7702 delegate.

        Returns:
            Optional[`:class:~ape.contracts.ContractInstance`]:
                The contract instance of the delegate contract (if available).
        """
        return self.chain_manager.get_delegate(self.address)

    @cached_property
    def history(self) -> "AccountHistory":
        """
        The list of transactions that this account has made on the current chain.
        """
        return self.chain_manager.history[self.address]

    def as_transaction(self, **kwargs) -> "TransactionAPI":
        sign = kwargs.pop("sign", False)
        converted_kwargs = self.conversion_manager.convert_method_kwargs(kwargs)
        tx = self.provider.network.ecosystem.create_transaction(
            receiver=self.address, **converted_kwargs
        )
        if sender := kwargs.get("sender"):
            if hasattr(sender, "prepare_transaction"):
                prepared = sender.prepare_transaction(tx)
                return (sender.sign_transaction(prepared) or prepared) if sign else prepared

        return tx

    def estimate_gas_cost(self, **kwargs) -> int:
        txn = self.as_transaction(**kwargs)
        return self.provider.estimate_gas_cost(txn)

    def prepare_transaction(self, txn: "TransactionAPI", **kwargs) -> "TransactionAPI":
        """
        Set default values on a transaction.

        Raises:
            :class:`~ape.exceptions.AccountsError`: When the account cannot afford the transaction
              or the nonce is invalid.
            :class:`~ape.exceptions.TransactionError`: When given negative required confirmations.

        Args:
            txn (:class:`~ape.api.transactions.TransactionAPI`): The transaction to prepare.
            **kwargs: Sub-classes, such as :class:`~ape.api.accounts.AccountAPI`, use additional kwargs.

        Returns:
            :class:`~ape.api.transactions.TransactionAPI`
        """

        # NOTE: Allow overriding nonce, assume user understands what this does
        if txn.nonce is None:
            txn.nonce = self.nonce
        elif txn.nonce < self.nonce:
            raise AccountsError("Invalid nonce, will not publish.")

        txn = self.provider.prepare_transaction(txn)

        if (
            txn.sender not in self.account_manager.test_accounts._impersonated_accounts
            and txn.total_transfer_value > self.balance
        ):
            raise AccountsError(
                f"Transfer value meets or exceeds account balance "
                f"for account '{self.address}' on chain '{self.chain_manager.chain_id}' "
                f"using provider '{self.provider.name}'.\n"
                "Are you using the correct account / chain / provider combination?\n"
                f"(transfer_value={txn.total_transfer_value}, balance={self.balance})."
            )

        txn.sender = txn.sender or self.address
        return txn


class Address(BaseAddress):
    """
    A generic blockchain address.

    Typically, this is used when we do not know the contract type at a given address,
    or to refer to an EOA the user doesn't personally control.
    """

    def __init__(self, address: AddressType):
        self._address = address

    @property
    def address(self) -> AddressType:
        """
        The raw address type.

        Returns:
            :class:`~ape.types.address.AddressType`: An alias to
            `ChecksumAddress <https://eth-typing.readthedocs.io/en/latest/types.html#checksumaddress>`__.  # noqa: E501
        """
        return self._address
